"""
[CT] 仙人掌 (Cactus)：每个雷周围四格中恰有1或3个雷。
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch

class RuleCT(AbstractMinesRule):
    name = ["CT", "仙人掌", "Cactus"]
    doc = "每个雷周围四格中恰有1或3个雷"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="var"):
            var_list = board.batch(pos.neighbors(1), mode="var", drop_none=True)
            sum_var = model.NewIntVar(0, len(var_list), "[CT]")
            model.Add(sum(var_list) == sum_var)
            model.AddAllowedAssignments([sum_var], [[1], [3]]).OnlyEnforceIf([var, s])


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/08 11:12
# @Author  : Wu_RH
# @FileName: 2G.py
"""
[2G] 四连块 (Group)：所有四连通雷区域的面积为 4
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition


class Rule2G(AbstractMinesRule):
    name = ["2G", "四连块", "Group"]
    doc = "所有四连通雷区域的面积为 4"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        self.nei_values = []
        if data is None:
            self.nei_values = [tuple([1])]
            self.value = 4
            return
        nei_values = "1"
        if ";;" in data:
            nei_values = data.split(";;")[0]
            self.value = int(data.split(";;")[1])
            if self.value > 6:
                raise ValueError("连块数量禁止超过6")
        if nei_values == "":
            nei_values = "1"
        nei_values = nei_values.split(";")
        for nei_value in nei_values:
            if ":" in nei_value:
                self.nei_values.append(tuple([
                    int(nei_value.split(":")[0]),
                    int(nei_value.split(":")[1])
                ]))
            else:
                self.nei_values.append(tuple([int(nei_value)]))

    def nei_pos(self, pos: AbstractPosition):
        positions = []
        for nei_value in self.nei_values:
            if len(nei_value) == 1:
                positions.extend(
                    pos.neighbors(nei_value[0], nei_value[0])
                )
            elif len(nei_value) == 2:
                positions.extend(
                    pos.neighbors(nei_value[0], nei_value[1])
                )
        return positions

    def create_constraints(self, board: AbstractBoard, switch):
        model = board.get_model()
        s = switch.get(model, self)

        def dfs(_board: AbstractBoard, _valides: list, step=0, checked=None, _possible_list=None):
            if _possible_list is None:
                _possible_list = set()
            if checked is None:
                checked = []
            if step == self.value:
                _possible_list.add((tuple(sorted(set(_valides))), tuple(sorted(set(checked)))))
                return _possible_list
            for _pos in sorted(set(_valides)):
                if _pos in checked:
                    continue
                if board.get_type(_pos) == "C":
                    continue
                checked.append(_pos)
                _valides.remove(_pos)
                pos_list = []
                for __pos in self.nei_pos(_pos):
                    if __pos in checked:
                        continue
                    if __pos in _valides:
                        continue
                    if not board.in_bounds(__pos):
                        continue
                    _valides.append(__pos)
                    pos_list.append(__pos)
                dfs(_board, _valides, step + 1, checked, _possible_list)
                for __pos in pos_list:
                    _valides.remove(__pos)
                checked.remove(_pos)
                _valides.append(_pos)
            return _possible_list

        for pos, var in board("NF", mode="variable"):
            vaildes = [pos]

            tmp_list = []
            possible_list = dfs(board, vaildes)
            for vars_t, vars_f in possible_list:
                if any(t_pos in vars_f for t_pos in vars_t):
                    continue
                if any(f_pos in vars_t for f_pos in vars_f):
                    continue
                # print(vars_t)
                # print(vars_f)
                # print()
                vars_t = board.batch(vars_t, mode="variable")
                vars_f = board.batch(vars_f, mode="variable")
                tmp = model.NewBoolVar("tmp")
                model.Add(sum(vars_t) == 0).OnlyEnforceIf([tmp, s])
                model.AddBoolAnd(vars_f).OnlyEnforceIf([tmp, s])
                tmp_list.append(tmp)
            model.AddBoolOr(tmp_list).OnlyEnforceIf([var, s])

    def suggest_total(self, info: dict):

        def hard_constraint(m, total):
            m.AddModuloEquality(0, total, self.value)

        info["hard_fns"].append(hard_constraint)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
[1E'] 视差 (Eyesight')：线索表示纵向和横向的视野之差，箭头指示视野更长的方向
"""
from typing import Dict

from minesweepervariants.utils.web_template import Number, StrWithArrow
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition

from ....utils.image_create import get_image, get_text, get_row, get_col, get_dummy


class Rule1E(AbstractClueRule):
    name = ["1E'", "E'", "视差", "Eyesight'"]
    doc = "线索表示纵向和横向的视野之差，箭头指示视野更长的方向"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            value = 0
            # 四方向的函数
            direction_funcs = [
                lambda _n: type(pos)(pos.x + _n, pos.y, pos.board_key),  # 右
                lambda _n: type(pos)(pos.x - _n, pos.y, pos.board_key),  # 左
                lambda _n: type(pos)(pos.x, pos.y + _n, pos.board_key),  # 上
                lambda _n: type(pos)(pos.x, pos.y - _n, pos.board_key)   # 下
            ]

            for fn in direction_funcs[:2]:  # 只计算横向
                n = 1
                while True:
                    next_pos = fn(n)
                    if not board.in_bounds(next_pos):
                        break
                    if board.get_type(next_pos) == "F":  # 遇到雷，视线被阻挡
                        break
                    value += 1
                    n += 1

            for fn in direction_funcs[2:]:  # 只计算纵向
                n = 1
                while True:
                    next_pos = fn(n)
                    if not board.in_bounds(next_pos):
                        break
                    if board.get_type(next_pos) == "F":  # 遇到雷，视线被阻挡
                        break
                    value -= 1
                    n += 1

            obj = Value1E(pos, bytes([value + 128]))
            board.set_value(pos, obj)
        return board


class Value1E(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.value = self.value - 128
        self.pos = pos

    def __repr__(self):
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        positions = []
        for i in [
            (1, 0), (-1, 0),
            (0, 1), (0, -1),
        ]:
            n = 0
            while board.get_type(pos := self.pos.shift(i[0] * n, i[1] * n)) not in "F":
                n += 1
                positions.append(pos)
        return positions

    @classmethod
    def type(cls) -> bytes:
        return Rule1E.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value+128])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        direction_funcs = [
            lambda n: type(self.pos)(self.pos.x + n, self.pos.y, self.pos.board_key),  # 右
            lambda n: type(self.pos)(self.pos.x - n, self.pos.y, self.pos.board_key),  # 左
            lambda n: type(self.pos)(self.pos.x, self.pos.y + n, self.pos.board_key),  # 上
            lambda n: type(self.pos)(self.pos.x, self.pos.y - n, self.pos.board_key)   # 下
        ]

        def max_steps(fn):
            n = 1
            while True:
                p = fn(n)
                if not board.in_bounds(p):
                    return n - 1
                if board.get_variable(p) is None:
                    return n - 1
                n += 1

        def collect_dir(fn, steps):
            _t_positions = []
            if steps == 0:
                p_block = fn(1)
                f_var = board.get_variable(p_block) if board.in_bounds(p_block) else None
                return _t_positions, f_var, True

            for k in range(1, steps + 1):
                p = fn(k)
                if not board.in_bounds(p):
                    return [], None, False
                var = board.get_variable(p)
                if var is None:
                    return [], None, False
                _t_positions.append(p)

            p_block = fn(steps + 1)
            f_var = board.get_variable(p_block) if board.in_bounds(p_block) else None
            return _t_positions, f_var, True

        max_right = max_steps(direction_funcs[0])
        max_left = max_steps(direction_funcs[1])
        max_up = max_steps(direction_funcs[2])
        max_down = max_steps(direction_funcs[3])

        possible_list = []  # 每项 (set_of_T_positions, list_of_F_vars_or_None)

        max_steps_list = [max_right, max_left, max_up, max_down]

        possible_list = []

        def enum_counts(idx: int, counts: list[int], accum_T: list, accum_F: list):
            # 计算当前已确定的 delta 与剩余方向的可达范围
            current_delta = (counts[0] + counts[1]) - (counts[2] + counts[3])
            # 计算剩余最大可增加的横向与纵向
            horiz_remain = 0
            vert_remain = 0
            for j in range(idx, 4):
                if j < 2:
                    horiz_remain += max_steps_list[j]
                else:
                    vert_remain += max_steps_list[j]

            min_possible = current_delta - vert_remain
            max_possible = current_delta + horiz_remain
            if not (min_possible <= self.value <= max_possible):
                return

            if idx == 4:
                if (counts[0] + counts[1]) - (counts[2] + counts[3]) == self.value:
                    # 复制 accum_F，因为里面可能包含 None
                    possible_list.append((set(accum_T), list(accum_F)))
                return

            fn = direction_funcs[idx]
            max_n = max_steps_list[idx]
            for steps in range(0, max_n + 1):
                t_pos, f_var, ok = collect_dir(fn, steps)
                if not ok:
                    continue

                # push
                added = len(t_pos)
                accum_T.extend(t_pos)
                accum_F.append(f_var)
                counts[idx] = steps

                enum_counts(idx + 1, counts, accum_T, accum_F)

                # pop
                for _ in range(added):
                    accum_T.pop()
                accum_F.pop()
                counts[idx] = 0

        enum_counts(0, [0, 0, 0, 0], [], [])

        tmp_list = []
        for t_positions, f_vars in possible_list:
            vars_t = board.batch(t_positions, mode="variable") if t_positions else []
            vars_f = [v for v in f_vars if v is not None]

            tmp = model.NewBoolVar(f"tmp_1E_{self.pos.x}_{self.pos.y}_{len(tmp_list)}")
            # 当 tmp 和 线索开关 s 同时成立时，T 位置均为非雷（sum == 0）
            model.Add(sum(vars_t) == 0).OnlyEnforceIf([tmp, s])
            # 阻挡位置（若有变量）全部为雷
            if vars_f:
                model.AddBoolAnd(vars_f).OnlyEnforceIf([tmp, s])
            tmp_list.append(tmp)

        if tmp_list:
            model.AddBoolOr(tmp_list).OnlyEnforceIf(s)

    def web_component(self, board) -> Dict:
        if self.value == 0:
            return Number(0)
        if self.value < 0:
            return get_col(
                get_image(
                    "double_arrow",
                    image_height=0.4,
                ),
                get_dummy(height=-0.1),
                get_text(str(-self.value))
            )
        if self.value > 0:
            return get_row(
                get_dummy(width=0.15),
                get_image(
                    "double_arrow",
                    style="transform: rotate(90deg);"
                ),
                get_dummy(width=-0.15),
                get_text(str(self.value)),
                get_dummy(width=0.15),
            )
    def web_component(self, board) -> Dict:
        if self.value == 0:
            return Number(0)
        if self.value < 0:
            return StrWithArrow(str(-self.value), "left_right")
        if self.value > 0:
            return StrWithArrow(str(self.value), "up_down")


    def compose(self, board):
        if self.value == 0:
            return super().compose(board)
        if self.value < 0:
            return get_col(
                get_image(
                    "double_horizontal_arrow",
                    image_height=0.4,
                ),
                get_dummy(height=-0.1),
                get_text(str(-self.value))
            )
        if self.value > 0:
            return get_row(
                    get_dummy(width=0.15),
                    get_image("double_vertical_arrow", ),
                    get_dummy(width=-0.15),
                    get_text(str(self.value)),
                    get_dummy(width=0.15),
            )



import asyncio
import sys
import threading
from pathlib import Path

from minesweepervariants.server.model import Model
from minesweepervariants.server.multiplayer import MPModel

from .config import HOT_RELOAD, MULTIPLAYER

import minesweepervariants

if HOT_RELOAD:
    try:
        import jurigged

        path = minesweepervariants.__package__
        if path is None:
            path = "."
        jurigged.watch(path)
    except ImportError:
        print("jurigged not installed, hot-reload disabled")

from minesweepervariants.utils import tool

import waitress

from .router import create_app
from .session import DataStore, SessionManager


async def main():
    print("Initializing database...")
    db = DataStore("session.json")
    await db.load()
    print("Database initialized.")

    _Model = MPModel if MULTIPLAYER else Model

    sm = SessionManager(db, _Model)
    app = create_app(sm, _Model)

    tool.LOGGER = None
    tool.get_logger(log_lv="DEBUG")
    port = int(sys.argv[1] if len(sys.argv) == 2 else "5050")
    host = "0.0.0.0"

    print(f"server start at {host}:{port}")
    threading.Thread(target=waitress.serve, args=(app,), kwargs={"host": host, "port": port}).start()

    await db.start()


asyncio.run(main())



from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition

from .connect import connect

class Rule1S(AbstractMinesRule):
    name = ["1S''", "S''", "传送门蛇", "Portal Snake"]
    doc = "所有雷构成一条蛇。蛇是一条宽度为 1 的四连通路径，不存在分叉、环、交叉。题板的上下左右视为连通。"

    def portalNeighbor(self, pos : AbstractPosition, board: AbstractBoard):
        boundary = board.boundary(pos.board_key)
        neighbors = pos.neighbors(1)
        result = []
        for p in neighbors:
            if board.in_bounds(p):
                result.append(p)
        if pos.x == 0:
            result.append(board.get_pos(boundary.x, pos.y))
        if pos.x == boundary.x:
            result.append(board.get_pos(0, pos.y))
        if pos.y == 0:
            result.append(board.get_pos(pos.x, boundary.y))
        if pos.y == boundary.y:
            result.append(board.get_pos(pos.x, 0))
        return result

    def create_constraints(self, board, switch):
        model = board.get_model()
        s = switch.get(model, self)

        connect(model, board, s, connect_value=1, nei_value=lambda pos: self.portalNeighbor(pos, board))

        tmp_list = []
        for pos, var in board(mode="variable"):
            tmp_bool = model.NewBoolVar(f"[1S''] {pos}")
            var_list = board.batch(self.portalNeighbor(pos, board), mode="variable", drop_none=True)
            model.Add(sum(var_list) == 1).OnlyEnforceIf([var, tmp_bool, s])
            model.Add(sum(var_list) == 2).OnlyEnforceIf([var, tmp_bool.Not(), s])
            tmp_list.append(tmp_bool)
        model.Add(sum(tmp_list) == 2).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:35
# @Author  : Wu_RH
# @FileName: 1U.py
"""
[1U] 一元 (Unary)：所有雷不能与其他雷相邻
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1H(AbstractMinesRule):
    name = ["1U", "一元", "Unary"]
    doc = "所有雷不能与其他雷相邻"
    subrules = [
        [True, "[1U]一元"]
    ]

    def create_constraints(self, board: 'AbstractBoard', switch):
        if not self.subrules[0][0]:
            return
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="variable"):
            if board.in_bounds(pos.down()):
                model.Add(board.get_variable(pos.down()) == 0).OnlyEnforceIf([var, s])
            if board.in_bounds(pos.up()):
                model.Add(board.get_variable(pos.up()) == 0).OnlyEnforceIf([var, s])
            if board.in_bounds(pos.right()):
                model.Add(board.get_variable(pos.right()) == 0).OnlyEnforceIf([var, s])
            if board.in_bounds(pos.left()):
                model.Add(board.get_variable(pos.left()) == 0).OnlyEnforceIf([var, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

"""
[1X-] 残缺十字 (Pawn)：线索表示朝向一个方向的两个格子中的雷数，线索会标注出方向
"""
from typing import List, Dict

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.image_create import get_image, get_text, get_row, get_col, get_dummy
from ....utils.impl_obj import MINES_TAG, VALUE_QUESS
from ....utils.tool import get_random, get_logger


class Rule1X_(AbstractClueRule):
    name = ["1X-", "残缺十字", "Pawn"]
    doc = "线索表示朝向一个方向的两个格子中的雷数，线索会标注出方向"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        random = get_random()

        for pos, _ in board("N"):
            # 随机选择一个方向 (0:上, 1:右, 2:下, 3:左)
            direction = random.randint(0, 3)

            # 获取该方向上的两个格子
            target_positions = []
            if direction == 0:  # 上
                target_positions = [pos.up(), pos.up().up()]
            elif direction == 1:  # 右
                target_positions = [pos.right(), pos.right().right()]
            elif direction == 2:  # 下
                target_positions = [pos.down(), pos.down().down()]
            elif direction == 3:  # 左
                target_positions = [pos.left(), pos.left().left()]

            # 计算有效格子中的雷数
            mine_count = 0
            for target_pos in target_positions:
                if board.in_bounds(target_pos) and board.get_type(target_pos) == "F":
                    mine_count += 1

            board.set_value(pos, Value1X_(pos, count=mine_count, direction=direction))
            logger.debug(f"Set {pos} to 1X-[{mine_count}] direction {direction}")

        return board


class Value1X_(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, direction: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.count = code[0]
            self.direction = code[1]  # 0:上, 1:右, 2:下, 3:左
        else:
            # 直接初始化
            self.count = count
            self.direction = direction

        # 计算目标格子位置
        self.target_positions = []
        if self.direction == 0:  # 上
            self.target_positions = [self.pos.up(), self.pos.up().up()]
        elif self.direction == 1:  # 右
            self.target_positions = [self.pos.right(), self.pos.right().right()]
        elif self.direction == 2:  # 下
            self.target_positions = [self.pos.down(), self.pos.down().down()]
        elif self.direction == 3:  # 左
            self.target_positions = [self.pos.left(), self.pos.left().left()]

    def __repr__(self):
        direction_symbols = ['↑', '→', '↓', '←']
        return f"{self.count}{direction_symbols[self.direction]}"

    def __str__(self):
        direction_symbols = ['↑', '→', '↓', '←']
        return f"{self.count}{direction_symbols[self.direction]}"

    @classmethod
    def type(cls) -> bytes:
        return b'1X-'

    def code(self) -> bytes:
        return bytes([self.count, self.direction])

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        return self.target_positions

    def web_component(self, board) -> Dict:
        direction_images = ['up', 'right', 'down', 'left']

        if self.direction in [0, 2]:  # 上或下
            if self.count == 1:
                return get_row(
                        get_dummy(width=0.175),
                        get_text("1"),
                        get_image(direction_images[self.direction]),
                        get_dummy(width=0.175),
                    )

            return get_row(
                    get_text(str(self.count)),
                    get_image(direction_images[self.direction]),
                    spacing=-0.1,
                )
        else:  # 左或右
            return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        direction_images[self.direction],
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.count)),
                    get_dummy(height=0.2)
                )

    def compose(self, board) -> Dict:
        """生成可视化组件"""
        direction_images = ['up', 'right', 'down', 'left']

        if self.direction in [0, 2]:  # 上或下
            if self.count == 1:
                return get_row(
                        get_dummy(width=0.175),
                        get_text("1"),
                        get_image(direction_images[self.direction]),
                        get_dummy(width=0.175),
                    )

            return get_row(
                    get_text(str(self.count)),
                    get_image(direction_images[self.direction]),
                    spacing=-0.1,
                )
        else:  # 左或右
            return get_col(
                    get_dummy(height=0.1),
                    get_image(
                        direction_images[self.direction],
                        image_height=0.2,
                        image_width=0.7
                    ),
                    get_dummy(height=-0.05),
                    get_text(str(self.count)),
                    get_dummy(height=0.2)
                )

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        """逻辑推理"""
        # 收集有效的目标格子
        valid_targets = []
        type_dict = {"N": [], "F": []}

        for target_pos in self.target_positions:
            if board.in_bounds(target_pos):
                valid_targets.append(target_pos)
                t = board.get_type(target_pos)
                if t in ("", "C"):
                    continue
                type_dict[t].append(target_pos)

        if not valid_targets:
            return False

        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])

        if n_num == 0:
            return False

        # 如果雷数已满，剩余格子都是安全的
        if f_num == self.count:
            for pos in type_dict["N"]:
                board.set_value(pos, VALUE_QUESS)
            return True

        # 如果所有格子都必须是雷才能满足条件
        if f_num + n_num == self.count:
            for pos in type_dict["N"]:
                board.set_value(pos, MINES_TAG)
            return True

        return False

    def create_constraints(self, board: 'AbstractBoard', switch):
        """创建CP-SAT约束"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集有效目标格子的变量
        target_vars = []
        for target_pos in self.target_positions:
            if board.in_bounds(target_pos):
                var = board.get_variable(target_pos)
                target_vars.append(var)

        # 添加约束：目标格子中的雷数等于count
        if target_vars:
            model.Add(sum(target_vars) == self.count).OnlyEnforceIf(s)



import pkgutil
import importlib
import pathlib
from abc import ABC, abstractmethod


class AbstractDye(ABC):
    name = None

    def __init__(self, args):
        self.args = args

    @abstractmethod
    def dye(self, board):
        """染色函数"""


# 动态递归导入当前目录所有模块和包
def _auto_import_modules():
    current_pkg = __name__
    current_path = pathlib.Path(__file__).parent

    for finder, name, ispkg in pkgutil.walk_packages([str(current_path)], prefix=current_pkg + "."):
        importlib.import_module(name)


_auto_import_modules()


def get_dye(name: str) -> AbstractDye | None:
    name = name[1:] if name.startswith("@") else name
    args = ""
    if ":" in name:
        index = name.index(":")
        name, args = name[:index], name[index + 1:]

    for cls in AbstractDye.__subclasses__():
        if cls.name == name:
            return cls(args)

    raise ValueError(f"未知的染色规则[@{name[1:] if name.startswith('@') else name}]")


def get_all_dye():
    result = {}
    for cls in AbstractDye.__subclasses__():
        result[cls.name] = cls.__doc__
    return result



from importlib.metadata import version

__version__ = version("minesweepervariants")

def tuple_version() -> tuple[int, int, int]:
    ver = __version__.split(".")
    match len(ver):
        case 1:
            return (int(ver[0]), 0, 0)
        case 2:
            return (int(ver[0]), int(ver[1]), 0)
        case 3:
            return (int(ver[0]), int(ver[1]), int(ver[2]))
        case _:
            return (0, 0, 0)



from .scripts.generate_puzzle import main as puzzle
from .scripts.generate_game import main as puzzle_query
from .scripts.generate_test import main as test


from . import AbstractDye

class DyeSP(AbstractDye):
    name = "sp" # spiral
    __doc__ = "螺旋染色"

    def parse_int(self, str_value, default):
        if str_value is None:
            return default
        try:
            return int(str_value)
        except ValueError:
            return default

    def __init__(self, args):
        # if args:
        #     self.base_color = True
        # else:
        #     self.base_color = False # 背景色是非染色（第一圈染色）

        if not args:
            self.base_color = False # 背景色是非染色（第一圈染色）
            self.clockwise = True # 顺时针
        else:
            all_args = args.split(':')
            self.base_color = (self.parse_int(all_args[0], 0) != 0)
            self.clockwise = True if len(all_args) < 2 else (self.parse_int(all_args[1], 1) != 0)
    
    def dye(self, board):
        dye = self.base_color
        for key in board.get_interactive_keys():
            '''
            5x5: (0,0) (0,1) (0,2) (0,3) (0,4) (1,4) (2,4) (3,4) (3,3) (3,2) (3,1) (2,1) (2,2)
            '''
            n, m = board.get_config(key, 'size')
            for i in range(n):
                for j in range(m):
                    board.set_dyed(board.get_pos(i, j), dye) # 背景色
            dye = not dye # 正色

            top, bottom, left, right = 0, n - 1, 0, m - 1
            while top <= bottom and left <= right:
                if self.clockwise:
                    for j in range(left, right + 1): # (top, left) -> (top, right)
                        board.set_dyed(board.get_pos(top, j), dye)
                    for i in range(top + 1, bottom): # (top+1, right) -> (bottom-1, right)
                        board.set_dyed(board.get_pos(i, right), dye)
                    for j in range(right - 1, left, -1): # (bottom-1, right-1) -> (bottom-1, left+1)
                        board.set_dyed(board.get_pos(bottom-1, j), dye)
                    for i in range(bottom - 2, top + 1, -1): # (bottom-2, left+1) -> (top, left+1)
                        board.set_dyed(board.get_pos(i, left+1), dye)
                else:
                    for i in range(top, bottom + 1):
                        board.set_dyed(board.get_pos(i, left), dye)
                    for j in range(left + 1, right):
                        board.set_dyed(board.get_pos(bottom, j), dye)
                    for i in range(bottom - 1, top, -1):
                        board.set_dyed(board.get_pos(i, right-1), dye)
                    for j in range(right - 2, left + 1, -1):
                        board.set_dyed(board.get_pos(top+1, j), dye)
                top += 2
                bottom -= 2
                left += 2
                right -= 2





#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/15 11:59
# @Author  : Wu_RH
# @FileName: 3D1T'.py

"""
[3D1T']必三连: 雷必然处在横竖对角构成三连
"""

from .. import Abstract3DMinesRule
from .....abs.board import AbstractBoard
from .....utils.tool import get_logger


class Rule3D1Tp(Abstract3DMinesRule):
    name = ["3D1T'", "3DT'", "三维必三连"]
    doc = "雷必然处在横竖对角构成三连"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        logger = get_logger()

        # 初始化位置覆盖字典
        position_coverage = {pos: [] for pos, _ in board()}

        # 遍历棋盘上的每个位置
        for pos, _ in board():
            # 生成26个三维方向 (dx, dy, dz)，排除(0,0,0)
            directions = []
            for dx in (-1, 0, 1):
                for dy in (-1, 0, 1):
                    for dz in (-1, 0, 1):
                        if dx == 0 and dy == 0 and dz == 0:
                            continue
                        directions.append((dx, dy, dz))

            for dx, dy, dz in directions:
                positions = [pos]  # 起始位置

                # 计算偏移位置1：水平移动 + 垂直移动
                p1 = pos.shift(dx, dy)  # 水平移动
                if p1 is not None:
                    if dz == 1:
                        p1 = self.up(board, p1, 1)  # 向上移动
                    elif dz == -1:
                        p1 = self.down(board, p1, 1)  # 向下移动
                positions.append(p1)

                # 计算偏移位置2：继续同方向移动
                p2 = None
                if p1 is not None:
                    p2 = p1.shift(dx, dy)  # 水平移动
                    if p2 is not None:
                        if dz == 1:
                            p2 = self.up(board, p2, 1)  # 向上移动
                        elif dz == -1:
                            p2 = self.down(board, p2, 1)  # 向下移动
                positions.append(p2)

                # 检查所有位置是否有效
                if not all(p is not None and board.is_valid(p) for p in positions):
                    continue  # 跳过无效位置

                # 获取位置对应的变量
                var_list = [board.get_variable(p) for p in positions]

                # 创建三连组变量：当且仅当三个位置都是雷时为真
                b = model.NewBoolVar(f"triple_{pos}_{dx}_{dy}_{dz}")
                model.AddBoolAnd(var_list).OnlyEnforceIf([b, s])
                model.AddBoolOr([v.Not() for v in var_list]).OnlyEnforceIf([b.Not(), s])

                # 记录此三连组覆盖的所有位置
                for p in positions:
                    position_coverage[p].append(b)

        # 只对雷位置添加约束：必须属于至少一个三连组
        for pos, var in board(mode="variable"):
            coverage_list = position_coverage.get(pos, [])

            if coverage_list:  # 确保该位置有三连组覆盖
                # 雷 => 至少属于一个三连组
                model.AddBoolOr(coverage_list).OnlyEnforceIf([var, s])
                logger.trace(f"Pos {pos}:雷必须属于{len(coverage_list)}个三连组之一")
            else:
                # 无三连组覆盖的位置不能是雷
                model.Add(var == 0).OnlyEnforceIf(s)
                logger.warning(f"位置{pos}无三连组覆盖，强制为非雷")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/06/03 19:00
# @Author  : Wu_RH
# @FileName: F0.py

"""
[F]标准雷线索: 雷线索表示该格是一个雷
"""

from ....abs.Mrule import Rule0F as AbstractRule0F


class RuleF(AbstractRule0F):
    name = ["F", "雷"]
    doc = "线索表示该格是一个非雷"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/06/03 18:59
# @Author  : Wu_RH
# @FileName: R0.py

"""
[R]总雷数: 题板内的雷数量为一个整数
"""

from ....abs.Lrule import Rule0R


class RuleR(Rule0R):
    name = ["R", "总雷数"]
    doc = "有时你会需要用到总雷数来推理"



"""
[3R]引用：在第 X 行或第 X 列的格子的周围八格不能恰好有 X 个雷
"""

from ....abs.board import AbstractBoard
from ....abs.Lrule import AbstractMinesRule


class Rule1T(AbstractMinesRule):
    name = ["3R", "引用", "Reference"]
    doc = "在第 X 行或第 X 列的格子的周围八格不能恰好有 X 个雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, _ in board():
            var_list = board.batch(pos.neighbors(2), mode="variable", drop_none=True)
            model.Add(sum(var_list) != pos.x + 1).OnlyEnforceIf(s)
            model.Add(sum(var_list) != pos.y + 1).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 09:33
# @Author  : Wu_RH
# @FileName: 1N.py
"""
[4V1N] 负雷 (Negative)：线索表示数字是两个题板中相同位置的其中一个3x3范围内染色格与非染色格的雷数差
"""
from .....abs.Rrule import AbstractClueRule, AbstractClueValue
from .....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD
from .....utils.impl_obj import VALUE_QUESS, MINES_TAG
from .....utils.tool import get_random

from . import BOARD_NAME_4V


class Rule1N(AbstractClueRule):
    name = ["4V1N", "负雷映射"]
    doc = "线索表示数字是两个题板中相同位置的其中一个3x3范围内染色格与非染色格的雷数差"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        size = (board.boundary().x + 1, board.boundary().y + 1)
        board.generate_board(BOARD_NAME_4V, size)
        board.set_config(BOARD_NAME_4V, "interactive", True)
        board.set_config(BOARD_NAME_4V, "row_col", True)
        board.set_config(BOARD_NAME_4V, "VALUE", VALUE_QUESS)
        board.set_config(BOARD_NAME_4V, "MINES", MINES_TAG)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        for pos, _ in board("N"):
            values = [-1, -1]
            pos.board_key = MASTER_BOARD
            values[0] = abs(sum(board.get_type(_pos) == "F" if
                            board.get_dyed(_pos)
                            else -(board.get_type(_pos) == "F")
                            for _pos in pos.neighbors(0, 2)))
            pos.board_key = BOARD_NAME_4V
            values[1] = abs(sum(board.get_type(_pos) == "F" if
                            board.get_dyed(_pos)
                            else -(board.get_type(_pos) == "F")
                            for _pos in pos.neighbors(0, 2)))
            r_value = 0 if random.random() > 0.7 else 1
            pos.board_key = MASTER_BOARD
            if board.get_type(pos) != "F":
                obj = Value1N(pos=pos, code=bytes([values[r_value]]))
                board.set_value(pos, obj)
            pos.board_key = BOARD_NAME_4V
            if board.get_type(pos) != "F":
                obj = Value1N(pos=pos, code=bytes([values[1 - r_value]]))
                board.set_value(pos, obj)

        return board


class Value1N(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.neighbors_list = []
        for key in [MASTER_BOARD, BOARD_NAME_4V]:
            _pos = pos.clone()
            _pos.board_key = key
            self.neighbors_list.append(_pos.neighbors(0, 2))
        self.value = code[0]
        self.pos = pos

    def __repr__(self) -> str:
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors_list[0] + self.neighbors_list[1]

    @classmethod
    def type(cls) -> bytes:
        return Rule1N.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        a = model.NewBoolVar("abs_diff")
        b = model.NewBoolVar("abs_diff")

        nei_a_d = [_pos for _pos in self.neighbors_list[0] if board.get_dyed(_pos)]
        nei_a_n = [_pos for _pos in self.neighbors_list[0] if not board.get_dyed(_pos)]
        nei_b_d = [_pos for _pos in self.neighbors_list[1] if board.get_dyed(_pos)]
        nei_b_n = [_pos for _pos in self.neighbors_list[1] if not board.get_dyed(_pos)]

        vars_a_d = board.batch(nei_a_d, mode="variable", drop_none=True)
        vars_a_n = board.batch(nei_a_n, mode="variable", drop_none=True)
        vars_b_d = board.batch(nei_b_d, mode="variable", drop_none=True)
        vars_b_n = board.batch(nei_b_n, mode="variable", drop_none=True)

        diff_a = sum(vars_a_d) - sum(vars_a_n)
        max_abs_a = len(vars_a_d) + len(vars_a_n)
        abs_diff_a = model.NewIntVar(0, max_abs_a, "abs_diff")

        diff_b = sum(vars_b_d) - sum(vars_b_n)
        max_abs_b = len(vars_b_d) + len(vars_b_n)
        abs_diff_b = model.NewIntVar(0, max_abs_b, "abs_diff")

        model.AddAbsEquality(abs_diff_a, diff_a)
        model.Add(abs_diff_a == self.value).OnlyEnforceIf([a, s])
        model.AddAbsEquality(abs_diff_b, diff_b)
        model.Add(abs_diff_b == self.value).OnlyEnforceIf([b, s])

        model.AddBoolOr([a, b]).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/15 11:59
# @Author  : Wu_RH
# @FileName: 1T'.py

"""
[1T']必三连: 雷必然处在横竖对角构成三连
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1Tp(AbstractMinesRule):
    name = ["1T'", "必三连"]
    doc = "雷必然处在横竖对角构成三连"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        # 存储每个位置所属的三连组变量
        position_coverage = {pos: [] for pos, _ in board()}

        # 生成所有可能的三连组（8个方向）
        for pos, _ in board():
            # 定义8个方向：右/左/下/上/右下/左下/右上/左上
            directions = [
                (0, 1), (0, -1), (1, 0), (-1, 0),  # 横竖
                (1, 1), (1, -1), (-1, 1), (-1, -1)  # 对角线
            ]

            for dx, dy in directions:
                positions = [
                    pos,
                    pos.shift(dx, dy),
                    pos.shift(2 * dx, 2 * dy)
                ]

                # 检查所有位置是否在棋盘内
                if all(board.is_valid(p) for p in positions):
                    var_list = [board.get_variable(p) for p in positions]

                    # 创建三连组变量：当且仅当三个位置都是雷时为真
                    b = model.NewBoolVar(f"triple_{pos}_{dx}_{dy}")
                    model.AddBoolAnd(var_list).OnlyEnforceIf([b, s])
                    model.AddBoolOr([v.Not() for v in var_list]).OnlyEnforceIf([b.Not(), s])

                    # 记录此三连组覆盖的所有位置
                    for p in positions:
                        position_coverage[p].append(b)

        # 只对雷位置添加约束：必须属于至少一个三连组
        for pos, var in board(mode="variable"):
            coverage_list = position_coverage[pos]

            if coverage_list:  # 确保该位置有三连组覆盖
                # 雷 => 至少属于一个三连组
                model.AddBoolOr(coverage_list).OnlyEnforceIf([var, s])
            else:
                # 无三连组覆盖的位置不能是雷
                model.Add(var == 0).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 09:43
# @Author  : Wu_RH
# @FileName: 1Q1L.py
"""
[QL]误差无方:误差线索比真实值大1或小1，如果线索处在2*2非雷框内，则它是误差线索，反之则是真实值。
"""

from typing import List

from ....abs.Rrule import AbstractClueValue, AbstractClueRule
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_random


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    return [
        a_pos if board.is_valid(a_pos) else None,
        b_pos if board.is_valid(b_pos) else None,
        c_pos if board.is_valid(c_pos) else None,
        d_pos if board.is_valid(d_pos) else None
    ]


class Rule1Q1L(AbstractClueRule):
    name = ["QL", "1Q1L", "误差无方"]
    doc = "误差线索比真实值大1或小1，如果线索处在2*2非雷框内，则它是误差线索，反之则是真实值。"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        for pos, _ in board("N"):
            value = len([_pos for _pos in pos.neighbors(2) if board.get_type(_pos) == "F"])
            if value == 0:
                random_value = 1
            elif value == 8:
                random_value = 7
            else:
                if random.random() > 0.5:
                    random_value = (value + 1)
                else:
                    random_value = (value - 1)
            flag = False
            for _pos in block(pos.down().right(), board):
                if _pos is None:
                    continue
                _block = block(_pos, board)
                if None in _block:
                    continue
                if board.batch(_block, "type").count("F") == 0:
                    flag = True
            if flag:
                board.set_value(pos, Value1Q1L(pos, count=random_value))
            else:
                board.set_value(pos, Value1Q1L(pos, count=value))
        return board

    def create_constraints(self, board: 'AbstractBoard', switch) -> bool:
        block_map = {}
        model = board.get_model()
        for pos, _ in board():
            t = model.NewBoolVar("t")
            block_vars = block(pos.down().right(), board)
            if None in block_vars:
                continue
            block_vars = board.batch(block_vars, "variable")
            model.Add(sum(block_vars) == 0).OnlyEnforceIf(t)
            model.Add(sum(block_vars) > 0).OnlyEnforceIf(t.Not())
            block_map[pos] = t
        for pos, obj in board("C"):
            if type(obj) is not Value1Q1L:
                continue
            var_list = []
            for _pos in block(pos, board):
                if _pos is None:
                    continue
                if _pos not in block_map:
                    continue
                var_list.append(block_map[_pos])
            obj.create_constraints_(board, var_list, switch)
        return True


class Value1Q1L(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            # 从字节码解码
            self.value = code[0]
        else:
            # 直接初始化
            self.value = count
        self.neighbor = self.pos.neighbors(2)

    def __repr__(self):
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule1Q1L.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints_(self, board: 'AbstractBoard', var_list: list, switch):
        """创建CP-SAT约束：周围雷数等于count"""
        model = board.get_model()
        s = switch.get(model, self)

        # 收集周围格子的布尔变量
        neighbor_vars = []
        for neighbor in self.neighbor:  # 8方向相邻格子
            if board.in_bounds(neighbor):
                var = board.get_variable(neighbor)
                neighbor_vars.append(var)

        # 添加约束：周围雷数等于count+-1
        if not neighbor_vars:
            return

        neighbor_sum = sum(neighbor_vars)
        # 两个布尔变量表示加和为 count + 1 或 count - 1
        b1 = model.NewBoolVar("[1Q1L]")
        b2 = model.NewBoolVar("[1Q1L]")
        b3 = model.NewBoolVar("[1Q1L]")

        model.Add(sum(var_list) == 0).OnlyEnforceIf([b3.Not(), s])
        model.Add(sum(var_list) > 0).OnlyEnforceIf([b3, s])

        # 将布尔变量与表达式绑定
        model.Add(neighbor_sum == self.value + 1).OnlyEnforceIf([b1, s])
        model.Add(neighbor_sum != self.value + 1).OnlyEnforceIf([b1.Not(), s])

        model.Add(neighbor_sum == self.value - 1).OnlyEnforceIf([b2, s])
        model.Add(neighbor_sum != self.value - 1).OnlyEnforceIf([b2.Not(), s])

        model.Add(neighbor_sum == self.value).OnlyEnforceIf([b3.Not(), s])
        model.Add(neighbor_sum != self.value).OnlyEnforceIf([b3, s])

        model.Add(sum([b1, b2, b3.Not()]) == 1).OnlyEnforceIf(s)



#!/usr/bin/env python3

"""
[1X*] 王后 (Queen)：线索数表示斜向和横纵所有格子中的雷数
"""
from .CQ import Rule1XStar as AbstractClueRule


class Rule1XStar(AbstractClueRule):
    name = ["1X*", "王后", "Queen"]
    doc = "线索数表示斜向和横纵所有格子中的雷数"



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/15 16:34
# @Author  : Wu_RH
# @FileName: 1H2H.py
"""
[1H2H]: 在每一行中, 要么都没有横向相邻的雷，要么都至少有一个横向相邻的雷。
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1H(AbstractMinesRule):
    name = ["1H2H", "横横向向"]
    doc = "在每一行中, 要么都没有横向相邻的雷，要么都至少有一个横向相邻的雷。"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for key in board.get_interactive_keys():
            pos_bound = board.boundary(key)
            for _pos in board.get_col_pos(pos_bound):
                tmp = model.NewBoolVar("tmp")
                for pos in board.get_row_pos(_pos):
                    var = board.get_variable(pos)
                    pos_vars = []
                    if board.in_bounds(pos.right()):
                        pos_vars.append(board.get_variable(pos.right()))
                    if board.in_bounds(pos.left()):
                        pos_vars.append(board.get_variable(pos.left()))
                    model.Add(sum(pos_vars) == 0).OnlyEnforceIf([var, tmp, s])
                    model.AddBoolOr(pos_vars).OnlyEnforceIf([var, tmp.Not(), s])

    @classmethod
    def method_choose(cls) -> int:
        return 1



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 16:38
# @Author  : Wu_RH
# @FileName: 2E'.py
"""
[2E'']互指: 如果线索X周围有N个雷 则另一个题板的X=N的格子必定为雷
"""
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG
from ....utils.tool import get_random, get_logger

from ....abs.Rrule import AbstractClueValue, AbstractClueRule
from ....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD

ALPHABET = "ABCDEFGHI"
NAME_2Epp = "2E''"


class Rule2Ep(AbstractClueRule):
    name = ["2E''", "互指"]
    doc = "如果线索X周围有N个雷 则另一个题板的X=N的格子必定为雷"

    def __init__(self, board: AbstractBoard, data=None):
        super().__init__()
        size = (board.boundary().x + 1, board.boundary().y + 1)
        board.set_config(MASTER_BOARD, "pos_label", True)
        board.generate_board(NAME_2Epp, size)
        board.set_config(NAME_2Epp, "pos_label", True)
        board.set_config(NAME_2Epp, "interactive", True)
        board.set_config(NAME_2Epp, "row_col", True)
        board.set_config(NAME_2Epp, "VALUE", VALUE_QUESS)
        board.set_config(NAME_2Epp, "MINES", MINES_TAG)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        logger = get_logger()
        for (key_a, key_b) in [
            (MASTER_BOARD, NAME_2Epp),
            (NAME_2Epp, MASTER_BOARD)
        ]:
            letter_map = {i: [] for i in range(9)}
            for pos, _ in board("F", key=key_a):
                if pos.y > 8:
                    continue
                letter = ALPHABET[pos.y]
                if pos.x not in letter_map:
                    letter_map[pos.x] = []
                letter_map[pos.x].append(letter)

            for pos, _ in board("N", key=key_b):
                positions = pos.neighbors(2)
                value = board.batch(positions, mode="type").count("F")
                if not letter_map[value]:
                    board.set_value(pos, VALUE_QUESS)
                    continue
                letter = random.choice(letter_map[value])
                obj = Value2Ep(pos, bytes([ALPHABET.index(letter)]))
                board.set_value(pos, obj)
                logger.debug(f"[2E''] put {letter}({value}) at {pos}")
        return board


class Value2Ep(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.value = code[0]    # 实际为第几列的字母
        self.neighbors = pos.neighbors(2)
        self.pos = pos

    def __repr__(self) -> str:
        return f"{ALPHABET[self.value]}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule2Ep.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        if self.pos.board_key == MASTER_BOARD:
            board_key = NAME_2Epp
        else:
            board_key = MASTER_BOARD
        pos = board.get_pos(0, self.value, board_key)
        line = board.get_col_pos(pos)
        # print(self.pos, self, pos)
        # print(self.neighbors)
        line = board.batch(line, mode="variable")
        sum_vers = sum(board.batch(self.neighbors, mode="variable", drop_none=True))
        for index in range(min(9, len(line))):
            var = line[index]
            model.Add(sum_vers != index).OnlyEnforceIf(var.Not()).OnlyEnforceIf(s)
            get_logger().trace(f"[2E'']: {self.pos} != {index} if {var} is 0")



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 03:33
# @Author  : Wu_RH
# @FileName: 2B.py
"""
[2B] 桥 (Bridge)：所有雷构成若干组桥。桥是从题版左边界八连通连接（水平或斜角连接）到右边界，宽度为 1、长度与题版宽度相等的一条路径
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2B(AbstractMinesRule):
    name = ["2B", "桥", "Bridge"]
    doc = "所有雷构成若干组桥。桥是从题版左边界八连通连接（水平或斜角连接）到右边界，宽度为 1、长度与题版宽度相等的一条路径"

    def create_constraints(self, board: 'AbstractBoard', switch):
        """
        约束建议提供:哈嘿袁
        """
        model = board.get_model()
        s = switch.get(model, self)

        boundary_pos = board.boundary()

        last_row = board.get_row_pos(boundary_pos)
        for index in range(len(last_row)):
            col_b = board.get_col_pos(last_row[index])
            col_var = []
            for index_t in range(len(col_b) - 1):
                tmp_t = model.NewBoolVar(f"tmp_t_{col_b[index_t]}_{col_b[index_t + 1]}")
                var_t_a = board.get_variable(col_b[index_t])
                var_t_b = board.get_variable(col_b[index_t + 1])
                model.AddBoolOr([var_t_a, var_t_b]).OnlyEnforceIf([tmp_t, s])
                model.AddBoolAnd([var_t_a.Not(), var_t_b.Not()]).OnlyEnforceIf([tmp_t.Not(), s])
                col_var.append(tmp_t)

            for index_a in range(len(col_b) - 1):
                for index_b in range(index_a + 1, len(col_b)):
                    if -1 < index - 1 < len(last_row):
                        col_c = board.batch(board.get_col_pos(last_row[index - 1]), mode="variable")
                        model.Add(
                            sum(col_c[index_a+1:index_b]) ==
                            sum(board.batch(col_b[index_a+2:index_b-1], mode="variable"))
                        ).OnlyEnforceIf(col_var[index_a+1:index_b-1] +
                                        [col_var[index_a].Not(), col_var[index_b-1].Not(), s])
                    if -1 < index + 1 < len(last_row):
                        col_a = board.batch(board.get_col_pos(last_row[index + 1]), mode="variable")
                        model.Add(
                            sum(col_a[index_a+1:index_b]) ==
                            sum(board.batch(col_b[index_a+2:index_b-1], mode="variable"))
                        ).OnlyEnforceIf(col_var[index_a+1:index_b-1] +
                                        [col_var[index_a].Not(), col_var[index_b-1].Not(), s])

            for index_t in range(1, len(col_b) - 1):
                if -1 < index - 1 < len(last_row):
                    col_c = board.batch(board.get_col_pos(last_row[index - 1]), mode="variable")
                    model.Add(
                        sum(col_c[:index_t+1]) ==
                        sum(board.batch(col_b[:index_t], mode="variable"))
                    ).OnlyEnforceIf(col_var[:index_t]+[col_var[index_t].Not(), s])
                if -1 < index + 1 < len(last_row):
                    col_a = board.batch(board.get_col_pos(last_row[index + 1]), mode="variable")
                    model.Add(
                        sum(col_a[:index_t+1]) ==
                        sum(board.batch(col_b[:index_t], mode="variable"))
                    ).OnlyEnforceIf(col_var[:index_t]+[col_var[index_t].Not(), s])

            for index_t in range(1, len(col_b) - 1):
                if -1 < index - 1 < len(last_row):
                    col_c = board.batch(board.get_col_pos(last_row[index - 1]), mode="variable")
                    model.Add(
                        sum(col_c[index_t:]) ==
                        sum(board.batch(col_b[index_t+1:], mode="variable"))
                    ).OnlyEnforceIf(col_var[index_t:] + [col_var[index_t-1].Not(), s])
                if -1 < index + 1 < len(last_row):
                    col_a = board.batch(board.get_col_pos(last_row[index + 1]), mode="variable")
                    model.Add(
                        sum(col_a[index_t:]) ==
                        sum(board.batch(col_b[index_t+1:], mode="variable"))
                    ).OnlyEnforceIf(col_var[index_t:] + [col_var[index_t-1].Not(), s])

        # 两个并排的非雷其上下两侧的雷数必然相同
        for pos in board.get_col_pos(boundary_pos):
            row = board.get_row_pos(pos)
            for index in range(len(row) - 1):
                pos_a, pos_b = row[index], row[index + 1]
                var_a, var_b = board.batch([pos_a, pos_b], mode="variable")
                line_a = board.get_col_pos(pos_a)
                line_a = line_a[:line_a.index(pos_a)]
                line_b = board.get_col_pos(pos_b)
                line_b = line_b[:line_b.index(pos_b)]
                vars_a = board.batch(line_a, mode="variable")
                vars_b = board.batch(line_b, mode="variable")
                model.Add(sum(vars_a) == sum(vars_b)).OnlyEnforceIf([var_a.Not(), var_b.Not(), s])

        # 列平衡
        row_positions = board.get_row_pos(boundary_pos)
        row_sums = [
            sum(board.get_variable(_pos) for _pos in board.get_col_pos(pos))
            for pos in row_positions
        ]
        # 所有 row_sums 相等
        for i in range(1, len(row_sums)):
            model.Add(row_sums[i] == row_sums[0]).OnlyEnforceIf(s)

    def suggest_total(self, info: dict):
        size_list = [info["size"][key] for key in info["interactive"]]

        def a(model, total):
            nonlocal size_list
            var_list = []
            for i, (height, width) in enumerate(size_list):
                n = model.NewIntVar(0, height * width, f"width_{i}")
                model.AddModuloEquality(0, n, width)
                var_list.append(n)
            model.Add(sum(var_list) == total)

        info["hard_fns"].append(a)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/15 11:59
# @Author  : Wu_RH
# @FileName: 1T.py

"""
[3D1T]无三连: 雷不能在横竖对角构成三连
"""

from .. import Abstract3DMinesRule
from .....abs.board import AbstractBoard


class Rule1T(Abstract3DMinesRule):
    name = ["3D1T", "无三连"]
    doc = "雷不能在横竖对角构成三连"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        for pos, _ in board():
            for positions in [
                [pos, pos.left(), pos.left().left()],
                [pos, pos.down(), pos.down().down()],
                [pos, pos.left().down(), pos.left().down().left().down()],
                [pos, pos.left().up(), pos.left().up().left().up()],
                [pos, self.up(board, pos), self.up(board, pos, 2)],
                [pos, self.up(board, pos.left()), self.up(board, pos.left(2), 2)],
                [pos, self.up(board, pos.up()), self.up(board, pos.up(2), 2)],
                [pos, self.up(board, pos.up().left()), self.up(board, pos.up(2).left(2), 2)],
            ]:
                if None in positions:
                    continue
                var_list = board.batch(positions, mode="variable")
                if True in [None is i for i in var_list]:
                    continue
                model.Add(sum(var_list) != 3).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 09:33
# @Author  : Wu_RH
# @FileName: 1N.py
"""
[1N] 负雷 (Negative)：线索表示 3x3 范围内染色格与非染色格的雷数差
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition


class Rule1N(AbstractClueRule):
    name = ["1N", "N", "负雷", "Negative"]
    doc = "线索表示 3x3 范围内染色格与非染色格的雷数差"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            value = sum(board.get_type(_pos) == "F" if
                        board.get_dyed(_pos)
                        else -(board.get_type(_pos) == "F")
                        for _pos in pos.neighbors(2))
            obj = Value1N(pos, bytes([abs(value)]))
            board.set_value(pos, obj)
        return board


class Value1N(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.nei = pos.neighbors(2)
        self.pos = pos

    def __repr__(self) -> str:
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.nei

    @classmethod
    def type(cls) -> bytes:
        return Rule1N.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        nei_a = [_pos for _pos in self.nei if board.get_dyed(_pos)]
        nei_b = [_pos for _pos in self.nei if not board.get_dyed(_pos)]

        vars_a = board.batch(nei_a, mode="variable", drop_none=True)
        vars_b = board.batch(nei_b, mode="variable", drop_none=True)

        diff = sum(vars_a) - sum(vars_b)

        # 估计最大绝对值可能为 len(vars_a) + len(vars_b)
        max_abs = len(vars_a) + len(vars_b)
        abs_diff = model.NewIntVar(0, max_abs, "abs_diff")

        model.AddAbsEquality(abs_diff, diff)
        model.Add(abs_diff == self.value).OnlyEnforceIf(s)



"""
[#]标签: 线索会有不同的规则, 每个线索的规则会单独标出
"""
from typing import List, Tuple, Optional

from ....abs.Rrule import AbstractClueRule
from ....abs.board import AbstractBoard
from ....abs.rule import AbstractRule
from ....utils.tool import get_random, get_logger
from ...impl_obj import get_rule
from ....utils.impl_obj import VALUE_QUESS

from . import AbstractClueSharp

class Rule0Sharp(AbstractClueSharp):
    name = ["0#", "标签", "Tag"]
    doc = "包含以下规则 [1L], [1N], [1E'], [1M1N], [1N1X], [2M], [2X']。线索值只有零。"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        rules_name = ["1L", "1N", "1E'", "1M1N", "1N1X", "2M", "2X'"]
        super().__init__(rules_name, board, data)
        self.rules = data

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        boards : list[AbstractBoard] = []
        random = get_random()
        for rule in self.shape_rule.rules:
            boards.append(rule.fill(board.clone()))
        for key in board.get_board_keys():
            for pos, _ in board("N", key=key):
                values = [_board.get_value(pos) for _board in boards if _board.get_value(pos).__repr__() == '0']
                if not values:
                    board.set_value(pos, VALUE_QUESS)
                else:
                    board.set_value(pos, random.choice(values))
        return board


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/08 06:30
# @Author  : Wu_RH
# @FileName: 1O.py
"""
[1O] 外部 (Outside)：非雷区域四连通；每个雷区域以四连通连接到题版边界
"""
from typing import List

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition
from .connect import connect


def block(a_pos: AbstractPosition, board: AbstractBoard) -> List[AbstractPosition]:
    b_pos = a_pos.up()
    c_pos = a_pos.left()
    d_pos = b_pos.left()
    if not board.in_bounds(d_pos):
        return []
    return [a_pos, b_pos, c_pos, d_pos]


class Rule1O(AbstractMinesRule):
    name = ["1O", "O", "外部", "Outside"]
    doc = "非雷区域四连通；每个雷区域以四连通连接到题版边界"

    def create_constraints(self, board: 'AbstractBoard', switch):
        positions_vars = [(pos, var) for pos, var in board("always", mode="variable")]
        model = board.get_model()
        s = switch.get(model, self)
        connect(
            model,
            board,
            ub=len(positions_vars),
            connect_value=0,
            nei_value=1,
            switch=s,
        )
        root_list = [model.NewBoolVar(f'root_{i}') for i in range(len(positions_vars))]
        for index, (pos, var) in enumerate(positions_vars):
            model.Add(var == 1).OnlyEnforceIf(root_list[index])
            flag = True
            for _pos in pos.neighbors(1):
                if not board.in_bounds(_pos):
                    flag = False
                    break
            if flag:
                model.Add(root_list[index] == 0)
                continue
            model.Add(root_list[index] == 1).OnlyEnforceIf(var)
            model.Add(root_list[index] == 0).OnlyEnforceIf(var.Not())
        connect(
            model,
            board,
            connect_value=1,
            nei_value=1,
            root_vars=root_list,
            ub=len(positions_vars),
            switch=s,
        )
        # 1O大定式
        for pos, _ in board():
            pos_list = block(pos, board)
            if not pos_list:
                continue
            a, b, c, d = board.batch(pos_list, mode="variable")
            model.AddBoolOr([a.Not(), b, c, d.Not()]).OnlyEnforceIf(s)  # 排除 1010
            model.AddBoolOr([a, b.Not(), c.Not(), d]).OnlyEnforceIf(s)  # 排除 0101



from . import AbstractDye


class DyeC(AbstractDye):
    name = "v" # vertical
    __doc__ = "竖直染色"

    def __init__(self, args):
        if args:
            self.offset = 1
        else:
            self.offset = 0

    def dye(self, board):
        dye = True
        for key in board.get_interactive_keys():
            dye = not dye
            for pos, _ in board(key=key):
                _dye = dye ^ ((pos.y + self.offset) % 2 > 0)
                board.set_dyed(pos, _dye)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/21 07:33
# @Author  : Wu_RH
# @FileName: 2S'.py
"""
[2S']分段': 每行连续雷长度不同
"""

from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch


class Rule(AbstractMinesRule):
    name = ["2S'", "分段'"]
    doc = "每行连续雷长度不同"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        for key in board.get_interactive_keys():
            pos_bound = board.boundary(key)
            for col_pos in board.get_col_pos(pos_bound):
                row = board.get_row_pos(col_pos)
                int_var_row = [model.NewIntVar(0, len(row), f"{pos}") for pos in row]
                var_row = board.batch(row, mode="var")

                model.Add(int_var_row[0] == var_row[0])

                for index in range(1, len(int_var_row)):
                    model.Add(
                        int_var_row[index] == int_var_row[index - 1] + 1
                    ).OnlyEnforceIf(var_row[index])
                    model.Add(int_var_row[index] == 0).OnlyEnforceIf(
                        var_row[index].Not()
                    )

                for index_a in range(len(int_var_row)):
                    for index_b in range(index_a + 1, len(int_var_row)):
                        model.Add(
                            int_var_row[index_a] !=
                            int_var_row[index_b]
                        ).OnlyEnforceIf([
                            var_row[index_a],
                            var_row[index_a + 1].Not(),
                            var_row[index_b], s
                        ] + ([] if len(row) == index_b + 1 else [
                            var_row[index_b + 1].Not()
                        ]))



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/12 20:44
# @Author  : Wu_RH
# @FileName: PF.py
"""
[PF]素因子（Prime Factor）:除0和1外的线索显示真实值的最大素因子
"""
from minesweepervariants.abs.Rrule import AbstractClueRule, AbstractClueValue
from minesweepervariants.abs.board import AbstractBoard, AbstractPosition
from minesweepervariants.impl.summon.solver import Switch

PRIME_MAP = {
    0: [0],
    1: [1],
    2: [2, 4, 8],
    3: [3, 6],
    5: [5],
    7: [7],
}


class RulePF(AbstractClueRule):
    name = ["PF", "素因子", "Prime Factor"]
    doc = "除0和1外的线索显示真实值的最大素因子"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            pos_list = pos.neighbors(2)
            value = board.batch(pos_list, mode="type").count("F")
            for _value in PRIME_MAP:
                if value in PRIME_MAP[_value]:
                    value = _value
                    continue
            obj = ValuePF(pos, bytes([value]))
            board[pos] = obj
        return board


class ValuePF(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos, code)
        self.value = code[0]

    def __repr__(self) -> str:
        return str(self.value)

    @classmethod
    def type(cls) -> bytes:
        return RulePF.name[0].encode()

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)

        var_list = board.batch(self.pos.neighbors(2), mode="var", drop_none=True)
        value_list = PRIME_MAP[self.value]
        tmp_list = []
        for value in value_list:
            tmp = model.NewBoolVar("tmp")
            model.Add(sum(var_list) == value).OnlyEnforceIf([s, tmp])
            tmp_list.append(tmp)
        model.AddBoolOr(tmp_list).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/03 05:23
# @Author  : Wu_RH
# @FileName: run.py
# @Version : 1.0.0
import shutil
import sys
import argparse
import textwrap

from minesweepervariants import puzzle_query
from minesweepervariants import puzzle
from minesweepervariants import test

from minesweepervariants.config.config import DEFAULT_CONFIG

# ==== 获取默认值 ====
defaults = {}
defaults.update(DEFAULT_CONFIG)

# ==== 参数解析 ====
parser = argparse.ArgumentParser(description="")

subparsers = parser.add_subparsers(dest='command', required=False)

parser_list = subparsers.add_parser('list', help='列出所有规则的文档说明')

parser.add_argument("-s", "--size", nargs="+",
                    help="纸笔的题板边长")
parser.add_argument("-t", "--total", type=int, default=defaults.get("total"),
                    help="总雷数")
parser.add_argument("-c", "--rules", nargs="+", default=[],
                    help="所有规则名")
parser.add_argument("-d", "--dye", default=defaults.get("dye"),
                    help="染色规则名称，如 @c")
parser.add_argument("-m", "--mask",  default=defaults.get("dye"),
                    help="染色规则名称，如 @c")
parser.add_argument("-r", "--used-r", action="store_true", default=defaults.get("used_r"),
                    help="推理是否加R")
parser.add_argument("-a", "--attempts", type=int, default=defaults.get("attempts"),
                    help="尝试生成题板次数")
parser.add_argument("-q", "--query", type=int, default=defaults.get("query"),
                    help="生成题板的至少有几线索推理")
parser.add_argument("-e", "--early-stop", action="store_true", default=False,
                    help="生成题板的时候达到指定线索数量推理的时候 直接退出 这会导致线索图不正确")
parser.add_argument("-v", "--vice-board", action="store_true", default=False,
                    help="启用后生成题板的时候可以删除副板的信息")
parser.add_argument("--test", action="store_true", default=False,
                    help="启用后将仅生成一份使用了规则的答案题板")
parser.add_argument("--seed", type=int, default=defaults.get("seed"),
                    help="随机种子")
parser.add_argument("--onseed",  action="store_true", default=False,
                    help="启用可循的种子来生成题板,速度会大幅降低")
parser.add_argument("--log-lv", default=defaults.get("log_lv"),
                    help="日志等级，如 DEBUG、INFO、WARNING")
parser.add_argument("--board-class", default=defaults.get("board_class"),
                    help="题板的类名/题板的名称 通常使用默认值即可")
parser_list.add_argument("--shell", action="store_true", default=False)
args = parser.parse_args()

# ==== 调用生成 ====


def print_with_indent(text, indent="\t"):
    width = shutil.get_terminal_size(fallback=(80, 24)).columns // 2
    # 减去缩进长度，避免超宽
    effective_width = width - len(indent.expandtabs())
    lines = text.splitlines()
    for line in lines:
        wrapped = textwrap.fill(line, width=effective_width,
                                initial_indent=indent,
                                subsequent_indent=indent)
        print(wrapped, flush=True)
    print()


if args.command == "list":
    from minesweepervariants.impl import rule
    rule_list = rule.get_all_rules()
    # print(rule_list)

    if args.shell:
        import random
        encode = "utf-8"
        split_symbol = ''.join([chr(random.randint(33, 126)) for _ in range(50)])
        result = split_symbol.encode(encode)
        for rule_line in ["L", "M", "R"]:
            for name in rule_list[rule_line].keys():
                if not rule_list[rule_line][name]['module_doc']:
                    unascii_name = [n for n in rule_list[rule_line][name]["names"] if not n.isascii()]
                    zh_name = unascii_name[0] if unascii_name else ""
                    part = f"[{name}]{zh_name}: " + rule_list[rule_line][name]["doc"]
                else:
                    part = rule_list[rule_line][name]['module_doc']
                result += part.encode(encode)
                result += split_symbol.encode(encode)  # 如果原 join 是用分隔符连接
            result += (split_symbol * 2).encode(encode)
        print("hex_start:" + result.hex() + ":hex_end", end="", flush=True)
        # print(result.decode(encode))
        sys.stdout.buffer.flush()
        sys.exit(0)

    for rule_line, rule_line_name in [
        ("L", "\n\n左线规则:"),
        ("M", "\n\n中线规则:"),
        ("R", "\n\n右线规则:"),
    ]:
        if rule_list[rule_line]:
            print(rule_line_name, flush=True)
        for name in rule_list[rule_line]:
            doc = rule_list[rule_line][name]["module_doc"]
            unascii_name = [n for n in rule_list[rule_line][name]["names"] if not n.isascii()]
            zh_name = unascii_name[0] if unascii_name else ""
            if not doc:
                doc = f"[{name}]{zh_name}: " + rule_list[rule_line][name]["doc"]
            print_with_indent(doc)

    sys.exit(0)

if args.size is None:
    parser.print_help()
    sys.exit(0)
else:
    if len(args.size) == 0:
        parser.print_help()
        sys.exit(0)
    elif len(args.size) == 1:
        size = (int(args.size[0]), int(args.size[0]))
    else:
        size = (int(args.size[0]), int(args.size[1]))

if args.seed != defaults.get("seed"):
    args.attempts = 1

for rule_name in args.rules:
    if "$" in rule_name:
        args.rules[args.rules.index(rule_name)] = rule_name.replace("$", "^")

if args.test:
    test(
        log_lv=args.log_lv,
        seed=args.seed,
        size=size,
        total=args.total,
        rules=args.rules,
        dye=args.dye,
        mask_dye=args.mask,
        board_class=args.board_class,
        unseed=not args.onseed,
    )
elif args.query == defaults.get("query"):
    puzzle(
        log_lv=args.log_lv,
        seed=args.seed,
        attempts=args.attempts,
        size=size,
        total=args.total,
        rules=args.rules,
        dye=args.dye,
        mask_dye=args.mask,
        drop_r=(not args.used_r),
        board_class=args.board_class,
        vice_board=args.vice_board,
        unseed=not args.onseed,
    )
else:
    puzzle_query(
        log_lv=args.log_lv,
        seed=args.seed,
        size=size,
        total=args.total,
        rules=args.rules,
        query=args.query,
        attempts=args.attempts,
        dye=args.dye,
        mask_dye=args.mask,
        drop_r=(not args.used_r),
        early_stop=args.early_stop,
        board_class=args.board_class,
        vice_board=args.vice_board,
        unseed=not args.onseed,
    )



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 12:14
# @Author  : xxx
# @FileName: 2F.py
"""
[2F]花田: 染色格中的雷周围四格内恰好有1个雷
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule2F(AbstractMinesRule):
    name = ["2F", "花田", "Flower"]
    doc = "染色格中的雷周围四格内恰好有1个雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, dye in board(mode="dye"):
            if not dye:
                continue
            positions = pos.neighbors(1)
            # sum(vals) 表示周围雷数
            sum_vals = sum(board.batch(positions, mode="variable", drop_none=True))
            val = board.get_variable(pos)
            # val 为1时，vals中必须有且仅有一个1
            # 约束：val=1 => sum(vals) == 1
            model.Add(sum_vals == 1).OnlyEnforceIf([val, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/28 16:28
# @Author  : Wu_RH
# @FileName: 3O.py
"""
[3O]横纵: 雷从四个方向中的任意一个连到题板外
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule3O(AbstractMinesRule):
    name = ["3O", "横纵"]
    doc = "雷从四个方向中的任意一个连到题板外"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="variable"):
            row = board.get_row_pos(pos)
            col = board.get_col_pos(pos)
            row_a = row[:row.index(pos)]
            row_b = row[row.index(pos)+1:]
            col_a = col[:col.index(pos)]
            col_b = col[col.index(pos)+1:]
            vars_row_a = board.batch(row_a, "variable")
            vars_row_b = board.batch(row_b, "variable")
            vars_col_a = board.batch(col_a, "variable")
            vars_col_b = board.batch(col_b, "variable")
            r_a = model.NewBoolVar("r_a")
            r_b = model.NewBoolVar("r_b")
            c_a = model.NewBoolVar("c_a")
            c_b = model.NewBoolVar("c_b")
            model.Add(sum(vars_row_a) == len(vars_row_a)).OnlyEnforceIf([r_a, s])
            model.Add(sum(vars_row_b) == len(vars_row_b)).OnlyEnforceIf([r_b, s])
            model.Add(sum(vars_col_a) == len(vars_col_a)).OnlyEnforceIf([c_a, s])
            model.Add(sum(vars_col_b) == len(vars_col_b)).OnlyEnforceIf([c_b, s])
            model.AddBoolOr([r_a, r_b, c_a, c_b]).OnlyEnforceIf([var, s])



from . import AbstractDye

class DyeW(AbstractDye):
    name = "w" # wrapsaround
    __doc__ = "回环染色"

    def __init__(self, args):
        if args:
            self.base_color = False
        else:
            self.base_color = True # 最外圈染色
    
    def dye(self, board):
        dye = self.base_color
        for key in board.get_interactive_keys():
            n, m = board.get_config(key, 'size')
            _dye = dye

            layers = (min(n, m) + 1) // 2
            for layer in range(layers):
                top, bottom = layer, n - layer - 1
                left, right = layer, m - layer - 1
                if top < bottom:
                    # top row
                    for j in range(left, right + 1):
                        board.set_dyed(board.get_pos(top, j), _dye)
                    # bottom row
                    for j in range(left, right + 1):
                        board.set_dyed(board.get_pos(bottom, j), _dye)
                if left < right:
                    # left column
                    for i in range(top, bottom + 1):
                        board.set_dyed(board.get_pos(i, left), _dye)
                    # right column
                    for i in range(top, bottom + 1):
                        board.set_dyed(board.get_pos(i, right), _dye)
                # 中心格
                if top == bottom and left == right:
                    board.set_dyed(board.get_pos(top, left), _dye)
                _dye = not _dye



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:13
# @Author  : Wu_RH
# @FileName: 1C.py
"""
[3D1C] 三维八连通 (Connected)：雷区域二十六连通
"""
from .. import Abstract3DMinesRule

from .connect import connect


class Rule1C(Abstract3DMinesRule):
    name = ["3D1C", "3DC", "二十六连通"]
    doc = "雷区域二十六连通"

    def create_constraints(self, board, switch):
        model = board.get_model()
        s = switch.get(model, self)

        connect(
            model=model,
            board=board,
            connect_value=1,
            nei_value=2,
            switch=s
        )



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/12 20:00
# @Author  : Wu_RH
# @FileName: 3C.py
"""
[3C']连通'（Connected'）: 每个雷周围四格中恰有两个雷
"""
from minesweepervariants.abs.Lrule import AbstractMinesRule
from minesweepervariants.abs.board import AbstractBoard
from minesweepervariants.impl.summon.solver import Switch


class Rule3C(AbstractMinesRule):
    name = ["3C'", "连通'", "Connected'"]
    doc = "每个雷周围四格中恰有两个雷"

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        for pos, var in board(mode="var"):
            var_list = board.batch(pos.neighbors(1), mode="var", drop_none=True)
            model.Add(sum(var_list) == 2).OnlyEnforceIf([var, s])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/04 10:33
# @Author  : Wu_RH
# @FileName: __init__.py
import heapq
from abc import ABC
from typing import List, Optional

from ....abs.Lrule import AbstractMinesRule
from ....abs.Mrule import AbstractMinesClueRule
from ....abs.Rrule import AbstractClueRule
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.impl_obj import VALUE_QUESS, MINES_TAG


class Abstract3DRule:
    def __init__(self, board: AbstractBoard, data: str = None):
        if data == "":
            return
        size = (board.boundary().x + 1, board.boundary().y + 1)
        for i in range((min(size) - 1) if data is None else int(data) - 1):
            key = f"{i + 2}"
            board.generate_board(key, size)
            board.set_config(key, "interactive", True)
            board.set_config(key, "row_col", True)
            board.set_config(key, "VALUE", VALUE_QUESS)
            board.set_config(key, "MINES", MINES_TAG)

    @staticmethod
    def pos_index(board: AbstractBoard, pos):
        return board.get_interactive_keys().index(pos.board_key)

    @classmethod
    def down(cls, board: AbstractBoard, pos: AbstractPosition, n=1):
        keys = board.get_interactive_keys()
        if pos.board_key not in keys:
            return None
        pos = pos.clone()
        index = keys.index(pos.board_key) - n
        if -1 < index < len(keys):
            pos.board_key = keys[index]
            return pos
        return None

    @classmethod
    def up(cls, board: AbstractBoard, pos: AbstractPosition, n=1):
        keys = board.get_interactive_keys()
        if pos.board_key not in keys:
            return None
        pos = pos.clone()
        index = keys.index(pos.board_key) + n
        if -1 < index < len(keys):
            pos.board_key = keys[index]
            return pos
        return None

    @staticmethod
    def pos_neighbors(board, pos: AbstractPosition, *args: int) -> List[AbstractPosition]:
        """
        按照欧几里得距离从小到大逐层扩散，筛选范围由距离平方控制（不包含当前位置）。

        调用方式（类似 range）：
            neighbors(end_layer)
                返回所有欧几里得距离 ≤ √end_layer 的位置（从第 1 层开始）。
            neighbors(start_layer, end_layer)
                返回所有欧几里得距离 ∈ [√start_layer, √end_layer] 的位置。

        :param board: 题板对象
        :param pos: 起始位置
        :param args: 一个或两个整数
            - 若提供一个参数 end_layer，视为从 √1 到 √end_layer。
            - 若提供两个参数 start_layer 和 end_layer，视为从 √start_layer 到 √end_layer。
            - 参数非法（数量不为 1 或 2，或值非法）时返回空列表。

        :return: 位置列表，按距离从近到远排序。
        """
        # 解析参数
        if len(args) == 1:
            start_sq = 1
            end_sq = args[0]
        elif len(args) == 2:
            start_sq = args[0]
            end_sq = args[1]
        else:
            return []

        # 检查参数有效性
        if end_sq < start_sq:
            return []

        # 定义移动函数（单位向量移动）
        def move_by_vector(
                _p: AbstractPosition, _dx: int, _dy: int, _dz: int
        ) -> Optional[AbstractPosition]:
            """按向量(dx, dy, dz)移动位置（每步移动1格）"""
            temp = _p

            # x轴移动
            if _dx != 0:
                if _dx > 0:
                    temp = temp.right(n=_dx)
                else:
                    temp = temp.left(n=-_dx)
                if temp is None:
                    return None

            # y轴移动（二维方向）
            if _dy != 0:
                if _dy > 0:
                    temp = temp.down(n=_dy)
                else:
                    temp = temp.up(n=-_dy)
                if temp is None:
                    return None

            # z轴移动（三维方向）
            if _dz != 0:
                if _dz > 0:
                    temp = Abstract3DRule.up(board, temp, n=_dz)
                else:
                    temp = Abstract3DRule.down(board, temp, n=-_dz)
                if temp is None:
                    return None

            return temp

        # 记录起点坐标
        x0, y0, z0 = pos.x, pos.y, Abstract3DRule.pos_index(board, pos)
        result = []

        # 生成26个三维方向向量（排除(0,0,0)）
        directions = []
        for dx in (-1, 0, 1):
            for dy in (-1, 0, 1):
                for dz in (-1, 0, 1):
                    if dx == 0 and dy == 0 and dz == 0:
                        continue
                    directions.append((dx, dy, dz))

        # 使用最小堆存储（距离平方, 位置）
        heap = []
        visited = set()
        visited.add((x0, y0, z0))  # 标记起点已访问

        # 初始化：加入所有相邻位置
        for dx, dy, dz in directions:
            neighbor = move_by_vector(pos, dx, dy, dz)
            if neighbor is None:
                continue
            n_coord = (neighbor.x, neighbor.y, Abstract3DRule.pos_index(board, neighbor))

            # 避免重复访问
            if n_coord in visited:
                continue
            visited.add(n_coord)

            # 计算欧几里得距离平方
            dist_sq = (neighbor.x - x0) ** 2 + (neighbor.y - y0) ** 2 + (
                        Abstract3DRule.pos_index(board, neighbor) - z0) ** 2
            heapq.heappush(heap, (dist_sq, neighbor))

        # 处理堆中的位置
        while heap:
            dist_sq, p = heapq.heappop(heap)

            # 超出范围时停止
            if dist_sq > end_sq:
                break

            # 在范围内则加入结果
            if dist_sq >= start_sq:
                result.append(p)

            # 扩展新位置
            for dx, dy, dz in directions:
                neighbor = move_by_vector(p, dx, dy, dz)
                if neighbor is None:
                    continue
                n_coord = (neighbor.x, neighbor.y, Abstract3DRule.pos_index(board, neighbor))

                if n_coord in visited:
                    continue
                visited.add(n_coord)

                # 计算新位置到起点的距离平方
                new_dist_sq = (neighbor.x - x0) ** 2 + (neighbor.y - y0) ** 2 + (
                            Abstract3DRule.pos_index(board, neighbor) - z0) ** 2
                heapq.heappush(heap, (new_dist_sq, neighbor))

        return result


class Abstract3DMinesRule(AbstractMinesRule, ABC, Abstract3DRule):
    def __init__(self, board: AbstractBoard, data: str = None):
        Abstract3DRule.__init__(self, board, data)
        super().__init__(board, data)


class Abstract3DClueRule(AbstractClueRule, ABC, Abstract3DRule):
    def __init__(self, board: AbstractBoard, data: str = None):
        Abstract3DRule.__init__(self, board, data)
        super().__init__(board, data)


class Abstract3DMinesClueRule(AbstractMinesClueRule, ABC, Abstract3DRule):
    def __init__(self, board: AbstractBoard, data: str = None):
        Abstract3DRule.__init__(self, board, data)
        super().__init__(board, data)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/29 05:15
# @Author  : Wu_RH
# @FileName: Quess.py
"""
[?]标准线索: 线索表示该格是一个非雷
"""

from ....abs.Rrule import AbstractClueRule, ValueQuess
from ....abs.board import AbstractBoard
from ....utils.impl_obj import VALUE_QUESS


class RuleQuess(AbstractClueRule):
    name = ["?", "问号"]
    doc = "线索表示该格是一个非雷"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        self.clear = data is not None

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            board.set_value(pos, VALUE_QUESS)
        return board

    def init_clear(self, board: 'AbstractBoard'):
        if self.clear:
            for pos, obj in board("C"):
                if obj is not VALUE_QUESS:
                    continue
                board.set_value(pos, None)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/10 09:52
# @Author  : xxx
# @FileName: c.py

from . import AbstractDye


class DyeC(AbstractDye):
    name = "3ds"
    __doc__ = "3D空心染色"

    def dye(self, board):
        dye = True
        keys = board.get_interactive_keys()
        keys = set(filter(lambda k: k.isdigit(), keys))
        last = max(keys)
        first = min(keys)
        for key in keys:
            boundary = board.boundary(key)
            if key in (first, last):
                continue
            for pos, _ in board(key=key):
                if pos.x == 0 or pos.y == 0 or pos.x == boundary.x or pos.y == boundary.y:
                    continue
                _dye = dye
                board.set_dyed(pos, _dye)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/04 07:39
# @Author  : Wu_RH
# @FileName: 4V.py
"""
[4V]2X'plus: 线索表示数字是两个题板中相同位置的其中一个范围中心3*3区域的雷总数
"""
from typing import List

from .....abs.Rrule import AbstractClueRule, AbstractClueValue
from .....abs.board import AbstractBoard, AbstractPosition, MASTER_BOARD
from .....utils.impl_obj import VALUE_QUESS, MINES_TAG
from .....utils.tool import get_random

from . import BOARD_NAME_4V


class Rule4V(AbstractClueRule):
    name = ["4V", "映射"]
    doc = "线索表示数字是两个题板中相同位置的其中一个范围中心3*3区域的雷总数"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        size = (board.boundary().x + 1, board.boundary().y + 1)
        board.generate_board(BOARD_NAME_4V, size)
        board.set_config(BOARD_NAME_4V, "interactive", True)
        board.set_config(BOARD_NAME_4V, "row_col", True)
        board.set_config(BOARD_NAME_4V, "VALUE", VALUE_QUESS)
        board.set_config(BOARD_NAME_4V, "MINES", MINES_TAG)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()

        for pos, _ in board():
            neighbors_list = []
            for _key in [MASTER_BOARD, BOARD_NAME_4V]:
                _pos = pos.clone()
                _pos.board_key = _key
                neighbors_list.append(_pos.neighbors(0, 2))
            values = [board.batch(positions, mode="type").count("F") for positions in neighbors_list]
            r_value = 0 if random.random() > 0.7 else 1
            _pos.board_key = MASTER_BOARD
            if board.get_type(_pos) != "F":
                obj = Value4V(pos=_pos, code=bytes([values[r_value]]))
                board.set_value(_pos, obj)
            _pos.board_key = BOARD_NAME_4V
            if board.get_type(_pos) != "F":
                obj = Value4V(pos=_pos, code=bytes([values[1 - r_value]]))
                board.set_value(_pos, obj)

        return board

    def suggest_total(self, info: dict):
        ub = 0
        for key in info["interactive"]:
            size = info["size"][key]
            ub += size[0] * size[1]
        info["soft_fn"](ub * 0.4)


class Value4V(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.neighbors_list = []
        for key in [MASTER_BOARD, BOARD_NAME_4V]:
            _pos = pos.clone()
            _pos.board_key = key
            self.neighbors_list.append(_pos.neighbors(0, 2))
        self.value = code[0]
        self.pos = pos

    @classmethod
    def method_choose(cls) -> int:
        return 1

    @classmethod
    def type(cls) -> bytes:
        return Rule4V.name[0].encode("ascii")

    def __repr__(self) -> str:
        return f"{self.value}"

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        neighbors1_type = board.batch(self.neighbors_list[0], mode="type")
        neighbors2_type = board.batch(self.neighbors_list[1], mode="type")
        if neighbors1_type.count("N") == 0:
            if neighbors1_type.count("F") == self.value:
                return self.neighbors_list[0]
            else:
                return self.neighbors_list[1]
        elif neighbors1_type.count("F") > self.value:
            return self.neighbors_list[1]
        elif neighbors1_type.count("N") + neighbors1_type.count("F") < self.value:
            return self.neighbors_list[1]
        if neighbors2_type.count("N") == 0:
            if neighbors2_type.count("F") == self.value:
                return self.neighbors_list[1]
            else:
                return self.neighbors_list[0]
        elif neighbors2_type.count("F") > self.value:
            return self.neighbors_list[0]
        elif neighbors2_type.count("F") + neighbors2_type.count("N") < self.value:
            return self.neighbors_list[0]
        return self.neighbors_list[0] + self.neighbors_list[1]

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self.pos)

        sum_var = []
        for neighbor in self.neighbors_list:
            var_list = board.batch(neighbor, mode="variable", drop_none=True)
            if var_list:
                b = model.NewBoolVar(f"[{Rule4V.name}]tmp")
                model.Add(sum(var_list) == self.value).OnlyEnforceIf([b, s])
                model.Add(sum(var_list) != self.value).OnlyEnforceIf([b.Not(), s])
                sum_var.append(b)
        model.AddBoolOr(sum_var).OnlyEnforceIf(s)

    def code(self) -> bytes:
        return bytes([self.value])



"""
[2K'] 钻石 (Diamond)：线索表示距离不超过 2 的范围内的雷数
"""

from minesweepervariants.utils.impl_obj import MINES_TAG, VALUE_QUESS
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger

class Rule2KPrime(AbstractClueRule):
    name = ["2K'", "钻石", "Diamond"]
    doc = "线索表示距离不超过 2 的范围内的雷数"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            neighbors = pos.neighbors(4) # 12 cells at distance 1, √2, 2
            value = sum(1 for n in neighbors if board.get_type(n) == "F")
            board.set_value(pos, Value2KPrime(pos, count=value))
            logger.debug(f"Set {pos} to 2K'[{value}]")
        return board

class Value2KPrime(AbstractClueValue):
    def __init__(self, pos: AbstractPosition, count: int = 0, code: bytes = None):
        super().__init__(pos, code)
        if code is not None:
            self.count = code[0]
        else:
            self.count = count
        self.neighbor = self.pos.neighbors(4)

    def __repr__(self):
        return f"{self.count}"

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbor

    @classmethod
    def type(cls) -> bytes:
        return Rule2KPrime.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.count])

    def deduce_cells(self, board: 'AbstractBoard') -> bool:
        type_dict = {"N": [], "F": []}
        for pos in self.neighbor:
            t = board.get_type(pos)
            if t in ("", "C"):
                continue
            type_dict[t].append(pos)
        n_num = len(type_dict["N"])
        f_num = len(type_dict["F"])
        if n_num == 0:
            return False
        if f_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, VALUE_QUESS)
            return True
        if f_num + n_num == self.count:
            for i in type_dict["N"]:
                board.set_value(i, MINES_TAG)
            return True
        return False
    
    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self.pos)
        neighbor_vars = []
        for neighbor in self.neighbor:
            var = board.get_variable(neighbor)
            if var is not None:
                neighbor_vars.append(var)
        if neighbor_vars:
            model.Add(sum(neighbor_vars) == self.count).OnlyEnforceIf(s)
            get_logger().trace(f"[2K'] {self.pos}: {self.count} constraint: {[str(n) for n in neighbor_vars]} == {self.count}")


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/11 14:03
# @Author  : xxx
# @FileName: a.py

from . import AbstractDye


class DyeA(AbstractDye):
    name = "a" # all
    __doc__ = "全盘染色"

    def dye(self, board):
        for key in board.get_interactive_keys():
            pos = board.boundary(key=key)
            for _pos in board.get_row_pos(pos):
                for __pos in board.get_col_pos(_pos):
                    board.set_dyed(__pos, True)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/01 07:30
# @Author  : Wu_RH
# @FileName: 2E.py
"""
[2E]加密: 线索被字母所取代，每个字母对应一个线索，且每个线索对应一个字母
"""
from typing import List, Dict

from minesweepervariants.utils.image_create import get_text
from minesweepervariants.utils.web_template import Number
from .....abs.board import AbstractBoard, AbstractPosition
from .....abs.Rrule import AbstractClueRule, AbstractClueValue
from .....utils.impl_obj import VALUE_QUESS, VALUE_CROSS, VALUE_CIRCLE
from .....utils.tool import get_random

NAME_2E = "2E"


class Rule2E(AbstractClueRule):
    name = ["2E", "加密"]
    doc = "线索被字母所取代，每个字母对应一个线索，且每个线索对应一个字母"

    def __init__(self, data=None, board: 'AbstractBoard' = None):
        super().__init__(board, data)
        pos = board.boundary()
        size = min(pos.x + 1, 9)
        board.generate_board(NAME_2E, (size, size))
        board.set_config(NAME_2E, "pos_label", True)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        shuffled_nums = [i for i in range(min(9, board.boundary().x + 1))]
        random.shuffle(shuffled_nums)
        for pos, _ in board("N"):
            count = board.batch(pos.neighbors(2), mode="type").count("F")
            if count not in shuffled_nums:
                board.set_value(pos, VALUE_QUESS)
            else:
                code = bytes([shuffled_nums[count]])
                board.set_value(pos, Value2E(pos, code))

        for x, y in enumerate(shuffled_nums):
            pos = board.get_pos(x, y, NAME_2E)
            board.set_value(pos, VALUE_CIRCLE)

        for pos, _ in board("N", key=NAME_2E):
            board.set_value(pos, VALUE_CROSS)

        return board

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        bound = board.boundary(key=NAME_2E)

        row = board.get_row_pos(bound)
        for pos in row:
            line = board.get_col_pos(pos)
            var = board.batch(line, mode="variable")
            model.Add(sum(var) == 1).OnlyEnforceIf(s)

        col = board.get_col_pos(bound)
        for pos in col:
            line = board.get_row_pos(pos)
            var = board.batch(line, mode="variable")
            model.Add(sum(var) == 1).OnlyEnforceIf(s)

    def init_clear(self, board: 'AbstractBoard'):
        for pos, _ in board(key=NAME_2E):
            board.set_value(pos, None)


class Value2E(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos
        self.neighbors = pos.neighbors(2)

    def __repr__(self) -> str:
        return "ABCDEFGHI"[self.value]

    def web_component(self, board) -> Dict:
        line = board.batch(board.get_col_pos(
            board.get_pos(0, self.value, NAME_2E)
        ), mode="type")
        if "F" in line:
            return Number(line.index("F"))
        return Number("ABCDEFGHI"[self.value])

    def compose(self, board) -> Dict:
        line = board.batch(board.get_col_pos(
            board.get_pos(0, self.value, NAME_2E)
        ), mode="type")
        if "F" in line:
            return get_text(str(line.index("F")))
        return get_text("ABCDEFGHI"[self.value])

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule2E.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        line = board.batch(board.get_col_pos(
            board.get_pos(0, self.value, NAME_2E)
        ), mode="variable")

        neighbors = board.batch(self.neighbors, mode="variable", drop_none=True)

        for index in range(len(line)):
            model.Add(sum(neighbors) == index).OnlyEnforceIf(line[index]).OnlyEnforceIf(s)
            model.Add(sum(neighbors) != index).OnlyEnforceIf(line[index].Not()).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/06/20 23:00
# @Author  : Wu_RH
# @FileName: 3B.py
"""
[3B]二进(Binary):无雷是0，有雷是1，线索代表每行（左起）和每列（上起）对应两个二进制数的按位异或值
"""
from typing import List

from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.tool import get_logger


def encode_bools_7bit(bools: list[bool]) -> bytes:
    # 编码头部：原始长度（用 4 字节 big-endian 表示）
    original_len = len(bools)
    length_bytes = [
        (original_len >> 24) & 0xFF,
        (original_len >> 16) & 0xFF,
        (original_len >> 8) & 0xFF,
        original_len & 0xFF
    ]

    # 编码主体：每 7 位布尔值 -> 1 字节（bit6~bit0，bit7固定为0）
    payload = []
    i = 0
    while i < len(bools):
        byte = 0
        for j in range(7):
            if i + j < len(bools) and bools[i + j]:
                byte |= 1 << (6 - j)
        payload.append(byte)  # bit7 默认为0
        i += 7

    return bytes(length_bytes + payload)


def decode_bools_7bit(data: bytes) -> list[bool]:
    if len(data) < 4:
        raise ValueError("数据不足4字节")

    # 解码头部长度信息（4字节 big-endian）
    original_len = (
            (data[0] << 24) |
            (data[1] << 16) |
            (data[2] << 8) |
            data[3]
    )

    # 解码主体数据
    bools = []
    for byte in data[4:]:
        for shift in range(6, -1, -1):
            bit = (byte >> shift) & 1
            bools.append(bool(bit))
            if len(bools) == original_len:
                return bools
    return bools[:original_len]  # 万一补了太多 false


class Rule3B(AbstractClueRule):
    name = ["3B", "二进", "Binary"]
    doc = "无雷是0，有雷是1，线索代表每行（左起）和每列（上起）对应两个二进制数的按位异或值"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        logger = get_logger()
        for pos, _ in board("N"):
            col = board.get_col_pos(pos)
            row = board.get_row_pos(pos)
            col = ["N" if i == "C" else i for i in board.batch(col, mode="type")]
            row = ["N" if i == "C" else i for i in board.batch(row, mode="type")]
            bools = [col[index] != row[index] for index in range(len(col))]
            obj = Value3B(pos, encode_bools_7bit(bools))
            board.set_value(pos, obj)
            logger.debug(f"Set {pos} [3B]{obj}")
        return board


class Value3B(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        super().__init__(pos)
        self.bools = decode_bools_7bit(code)
        self.__code = code

    def __repr__(self) -> str:
        result = 0
        for b in self.bools:
            result = (result << 1) | int(b)
        return f"{result}"

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition']:
        return board.get_row_pos(self.pos) + board.get_col_pos(self.pos)

    @classmethod
    def type(cls) -> bytes:
        return Rule3B.name[0].encode("ascii")

    def code(self) -> bytes:
        return self.__code

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        col = board.get_col_pos(self.pos)
        row = board.get_row_pos(self.pos)
        col = board.batch(col, mode="variable")
        row = board.batch(row, mode="variable")
        for index in range(len(self.bools)):
            if self.bools[index]:
                model.AddBoolOr([col[index].Not(), row[index].Not()]).OnlyEnforceIf(s)
                model.AddBoolOr([col[index], row[index]]).OnlyEnforceIf(s)
            else:
                model.AddBoolOr([col[index].Not(), row[index]]).OnlyEnforceIf(s)
                model.AddBoolOr([col[index], row[index].Not()]).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:13
# @Author  : Wu_RH
# @FileName: 1C.py
"""
[1C] 八连通 (Connected)：雷区域八连通
"""
from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.impl_obj import get_total

from .connect import connect


class Rule1C(AbstractMinesRule):
    name = ["1C", "C", "八连通", "Connected"]
    doc = "雷区域八连通"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        self.nei_values = []
        if data is None:
            self.nei_values = [tuple([1, 2])]
            return
        nei_values = data.split(";")
        for nei_value in nei_values:
            if ":" in nei_value:
                self.nei_values.append(tuple([
                    int(nei_value.split(":")[0]),
                    int(nei_value.split(":")[1])
                ]))
            else:
                self.nei_values.append(tuple([int(nei_value)]))

    def nei_pos(self, pos: AbstractPosition):
        positions = []
        for nei_value in self.nei_values:
            if len(nei_value) == 1:
                positions.extend(
                    pos.neighbors(nei_value[0], nei_value[0])
                )
            elif len(nei_value) == 2:
                positions.extend(
                    pos.neighbors(nei_value[0], nei_value[1])
                )
        return positions

    def create_constraints(self, board, switch):
        model = board.get_model()
        connect(
            model=model,
            board=board,
            connect_value=1,
            nei_value=self.nei_pos,
            switch=switch.get(model, self)
        )



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 15:57
# @Author  : QuirkyStorm7988
# @FileName: 1W'.py
"""
[1W'] 最长数墙 (Longest Wall)：线索表示 3x3 范围内最长的连续雷的长度
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition


def MineStatus_1W(clue: list) -> list[int]:
    """
    返回值：一个int列表，其中存的每一个int表示：
        一个二进制数，第i位（从低到高）表示从左上角开始顺时针旋转，第i个格子的雷情况（是雷->1，非雷->0）
        将这个二进制数转化为十进制存储到元素当中，如42(10) == 00101010(2)，即这个线索格的右上、右下、左下有雷
    """
    ans = []
    a = [0 for _ in range(8)]  # 决策列表

    def dfs(step: int):
        if step >= 8:  # 最终处理
            # 先写没有剪枝的
            test = []
            last = 0
            for i in range(8):
                if a[i]:
                    last += 1
                else:
                    if last != 0: test.append(last)
                    last = 0
            if last != 0: test.append(last)
            if a[-1] and a[0] and len(test) != 1:
                test[0] += test[-1]
                del test[-1]
            if not test: test = [0]
            test.sort()
            if test != clue: return None
            #
            status = 0
            for i in range(8):
                status += 2 ** i * a[i]
            if status not in ans:
                ans.append(status)
            # if a[:] not in ans:
            #     ans.append(a[:])
            return None
        a[step] = 0
        dfs(step + 1)
        a[step] = 1
        dfs(step + 1)
        return None

    dfs(0)
    return ans


class Rule1Wp(AbstractClueRule):
    name = ["1W'", "W'", "最长数墙", "Longest Wall"]
    doc = "线索表示 3x3 范围内最长的连续雷的长度"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            pos_list = [pos.right(), pos.right().down(), pos.down(), pos.left().down(),
                        pos.left(), pos.left().up(), pos.up(), pos.right().up()] * 2
            value = 0
            tmp = 0
            for _pos in pos_list:
                if board.get_type(_pos) == "F":
                    tmp += 1
                elif tmp != 0:
                    value = max(value, tmp)
                    tmp = 0
            if tmp > 8:
                value = 8
            obj = Value1Wp(pos, bytes([value]))
            board[pos] = obj
        return board


class Value1Wp(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.pos = pos

    def __repr__(self) -> str:
        return str(self.value)

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.pos.neighbors(2)

    @classmethod
    def type(cls) -> bytes:
        return Rule1Wp.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):

        def get_values(_value: int, _length=3):
            if _length <= 0:
                yield []
            if _value == 0:
                return []
            if _length > 0:
                for i in range(_value):
                    for j in get_values(i + 1, _length - 1):
                        yield [i + 1] + j

        model = board.get_model()
        s = switch.get(model, self)

        var_list = board.batch([
            self.pos.right(), self.pos.right().down(),
            self.pos.down(), self.pos.left().down(),
            self.pos.left(), self.pos.left().up(),
            self.pos.up(), self.pos.right().up()
        ], mode="variable")

        possible_list = [[]]
        values = []
        for length in range(3 - self.value // 2):
            if length == 0:
                values.extend(MineStatus_1W([self.value]))
            for k in get_values(3, _length=length + 1):
                values.extend(MineStatus_1W(([self.value] + k)[::-1]))

        for value in values:
            bool_list = [(value >> i) & 1 == 1 for i in reversed(range(8))]
            flag = False
            for index, var in enumerate(var_list):
                if var is None and bool_list[index]:
                    flag = True
                    break
                if var is None:
                    continue
                possible_list[-1].append(bool_list[index])
            if flag:
                possible_list.pop(-1)
            possible_list.append([])

        if any(v is None for v in var_list):
            var_list = [var for var in var_list if var is not None]
        possible_list.pop(-1)

        if possible_list:
            model.AddAllowedAssignments(var_list, possible_list).OnlyEnforceIf(s)
        else:
            model.Add(sum(var_list) == 0).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 16:43
# @Author  : Wu_RH
# @FileName: connect.py
from typing import List, Callable, Union

from ortools.sat.python import cp_model
from ortools.sat.python.cp_model import IntVar

from minesweepervariants.abs.board import AbstractBoard
from .. import Abstract3DRule


def connect(
        model: cp_model.CpModel,
        board: AbstractBoard,
        switch: IntVar,     # 连通性选择
        ub=False,  # 可达处的上限
        connect_value=1,  # 1=雷连通，0=非雷连通
        nei_value: Union[int, tuple, Callable] = 2,  # 1=四连通，2=八连通
        root_vars: List[IntVar] = None  # 允许提供根节点变量
):
    # 获取题板上所有位置及其对应的布尔变量
    positions_vars = [(pos, var) for pos, var in board("always", mode="variable")]
    if not positions_vars:
        return

    pos_list, var_list = zip(*positions_vars)
    n = len(pos_list)

    # 定义reach_vars整数变量
    reach_vars = [model.NewIntVar(0, (ub if ub else n + 1), f'reach_{i}') for i in range(n)]

    # 定义root_vars布尔变量
    if root_vars is None:
        root_vars = [model.NewBoolVar(f'root_{i}') for i in range(n)]
        model.Add(sum(root_vars) == 1).OnlyEnforceIf(switch)

    for i in range(n):
        # 根据connect_value决定连通对象
        if connect_value == 1:  # 雷连通
            model.AddImplication(root_vars[i], var_list[i]).OnlyEnforceIf(switch)
            model.Add(reach_vars[i] == 1).OnlyEnforceIf([root_vars[i], switch])
            model.Add(reach_vars[i] != 1).OnlyEnforceIf([root_vars[i].Not(), switch])
            model.Add(reach_vars[i] == 0).OnlyEnforceIf([var_list[i].Not(), switch])
        else:  # 非雷连通
            model.AddImplication(root_vars[i], var_list[i].Not()).OnlyEnforceIf(switch)
            model.Add(reach_vars[i] == 1).OnlyEnforceIf([root_vars[i], switch])
            model.Add(reach_vars[i] != 1).OnlyEnforceIf([root_vars[i].Not(), switch])
            model.Add(reach_vars[i] == 0).OnlyEnforceIf([var_list[i], switch])

    # 构造邻接列表（根据nei_value决定连通方式）
    adj = [[] for _ in range(n)]
    for i, pos_i in enumerate(pos_list):
        for j, pos_j in enumerate(pos_list):
            if i != j and board.in_bounds(pos_j):
                # 根据nei_value判断连通方式
                if callable(nei_value):
                    is_neighbor = pos_j in nei_value(pos_i)
                elif type(nei_value) is int:  # 四连通
                    is_neighbor = pos_j in Abstract3DRule.pos_neighbors(board, pos_i, nei_value)
                elif type(nei_value) is tuple:  # 四连通
                    is_neighbor = pos_j in Abstract3DRule.pos_neighbors(board, pos_i, nei_value[0], nei_value[1])
                else:  # 八连通
                    raise ValueError("")
                if is_neighbor:
                    adj[i].append(j)

    # 传播约束
    for i in range(n):
        # 条件判断根据connect_value变化
        if connect_value == 1:  # 雷连通
            cond = [var_list[i], root_vars[i].Not()]
        else:  # 非雷连通
            cond = [var_list[i].Not(), root_vars[i].Not()]

        possible_sources = []
        for j in adj[i]:
            tmp = model.NewBoolVar(f'path_{j}_to_{i}')
            model.Add(reach_vars[i] == reach_vars[j] + 1).OnlyEnforceIf([tmp, switch])

            # 根据connect_value决定传播条件
            if connect_value == 1:
                model.AddImplication(tmp, var_list[j]).OnlyEnforceIf(switch)
            else:
                model.AddImplication(tmp, var_list[j].Not()).OnlyEnforceIf(switch)

            is_reach_j_pos = model.NewBoolVar(f'is_reach_pos_{j}')
            model.Add(reach_vars[j] > 0).OnlyEnforceIf([is_reach_j_pos, switch])
            model.Add(reach_vars[j] == 0).OnlyEnforceIf([is_reach_j_pos.Not(), switch])
            model.AddImplication(tmp, is_reach_j_pos).OnlyEnforceIf(switch)

            possible_sources.append(tmp)

        if possible_sources:
            model.AddBoolOr(possible_sources).OnlyEnforceIf(cond + [switch])

    # 最终约束
    for i in range(n):
        if connect_value == 1:  # 雷连通
            model.Add(reach_vars[i] > 0).OnlyEnforceIf([var_list[i], switch])
            model.Add(reach_vars[i] == 0).OnlyEnforceIf([var_list[i].Not(), switch])
        else:  # 非雷连通
            model.Add(reach_vars[i] > 0).OnlyEnforceIf([var_list[i].Not(), switch])
            model.Add(reach_vars[i] == 0).OnlyEnforceIf([var_list[i], switch])



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/07 17:13
# @Author  : Wu_RH
# @FileName: 1A.py
"""
[1A] 无马步 (Anti-Knight)：所有雷的马步位置不能有雷
"""

from ....abs.Lrule import AbstractMinesRule
from ....abs.board import AbstractBoard


class Rule1A(AbstractMinesRule):
    name = ["1A", "A", "无马步", "Anti-Knight"]
    doc = "所有雷的马步位置不能有雷"

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        b = switch.get(model, self)
        for pos, var in board(mode="variable"):
            pos_list = [
                pos.left(2).up(1),
                pos.left(2).down(1),
                pos.down(2).left(1),
                pos.down(2).right(1)
            ]
            var_list = board.batch(pos_list, mode="variable", drop_none=True)
            for _var in var_list:
                model.AddBoolOr([_var.Not(), var.Not()]).OnlyEnforceIf(b)



from typing import Literal


def Number(n: int | float | str):
    return {
        'type': 'template',
        'style': '',
        'value': {
            'name': 'str',
            'value': n
        }
    }


def MultiNumber(n: list[int | float | str]):
    return {
        'type': 'template',
        'style': '',
        'value': {
            'name': 'multiStr',
            'value': n
        }
    }

def StrWithArrow(n: str, arrow: Literal['up', 'down', 'left', 'right', 'up_down', 'left_right']):
    return {
        'type': 'template',
        'style': '',
        'value': {
            'name': 'strWithArrow',
            'value': {
                'text': n,
                'arrow': arrow
            }
        }
    }


#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 03:20
# @Author  : Wu_RH
# @FileName: 2E^.py
"""
[2E^] 加密^ (Encrypted^)：线索被字母替代，每个数字与字母两两对应 [副版规则]
"""

from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....utils.impl_obj import VALUE_QUESS, VALUE_CROSS, VALUE_CIRCLE
from ....utils.tool import get_random

NAME_2Eq = "2E^"


class Rule2Eq(AbstractClueRule):
    name = ["2E^", "加密^", "Encrypted^"]

    def __init__(self, data=None, board: 'AbstractBoard' = None):
        super().__init__(board, data)
        pos = board.boundary()
        size = min(pos.x + 1, 9)
        board.generate_board(NAME_2Eq, (size, size))
        board.set_config(NAME_2Eq, "pos_label", True)

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        while True:
            shuffled_nums = [i // 2 for i in range(min(18, (board.boundary().x + 1) * 2))]
            number_map = {i: [] for i in range(min(9, (board.boundary().x + 1)))}
            for i in range(min(18, (board.boundary().x + 1) * 2)):
                tmp_list = []
                while shuffled_nums:
                    index = int(random.random() * len(shuffled_nums))
                    if shuffled_nums[index] in number_map[i // 2]:
                        tmp_list.append(shuffled_nums.pop(index))
                        continue
                    number_map[i // 2].append(shuffled_nums.pop(index))
                    break
                if tmp_list:
                    shuffled_nums.extend(tmp_list)
            if not shuffled_nums:
                break

        for pos, _ in board("N"):
            count = board.batch(pos.neighbors(2), mode="type").count("F")
            if count > board.boundary().x:
                board.set_value(pos, VALUE_QUESS)
            else:
                code = bytes([random.choice(number_map[count])])
                board.set_value(pos, Value2Eq(pos, code))

        for x in number_map:
            for y in number_map[x]:
                pos = board.get_pos(x, y, NAME_2Eq)
                board.set_value(pos, VALUE_CIRCLE)

        for pos, _ in board("N", key=NAME_2Eq):
            board.set_value(pos, VALUE_CROSS)

        return board

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        bound = board.boundary(key=NAME_2Eq)

        row = board.get_row_pos(bound)
        for pos in row:
            line = board.get_col_pos(pos)
            var = board.batch(line, mode="variable")
            model.Add(sum(var) == 2).OnlyEnforceIf(s)

        col = board.get_col_pos(bound)
        for pos in col:
            line = board.get_row_pos(pos)
            var = board.batch(line, mode="variable")
            model.Add(sum(var) == 2).OnlyEnforceIf(s)

    def init_clear(self, board: 'AbstractBoard'):
        for pos, _ in board(key=NAME_2Eq):
            board.set_value(pos, None)


class Value2Eq(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.neighbors = pos.neighbors(2)
        self.pos = pos

    def __repr__(self) -> str:
        return "ABCDEFGHI"[self.value]

    def high_light(self, board: 'AbstractBoard') -> list['AbstractPosition']:
        return self.neighbors

    @classmethod
    def type(cls) -> bytes:
        return Rule2Eq.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)

        line = board.batch(board.get_col_pos(
            board.get_pos(0, self.value, NAME_2Eq)
        ), mode="variable")

        neighbors = board.batch(self.neighbors, mode="variable", drop_none=True)

        tmp_vars = []

        for index in range(len(line)):
            tmp_var = model.NewBoolVar("tmp")
            model.Add(sum(neighbors) != index).OnlyEnforceIf([line[index].Not(), s])
            model.Add(sum(neighbors) == index).OnlyEnforceIf([line[index], tmp_var, s])
            tmp_vars.append(tmp_var)

        model.AddBoolOr(tmp_vars).OnlyEnforceIf(s)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-

# @Time    : 2025/06/03 01:58
# @Author  : Wu_RH
# @FileName: board.py

from typing import List, Union, Tuple, Any, Generator, TYPE_CHECKING
import heapq

from ortools.sat.python import cp_model
from ortools.sat.python.cp_model import IntVar

from ....utils.impl_obj import VALUE_QUESS, MINES_TAG
from ....utils.tool import get_logger
from ....abs.board import AbstractBoard, AbstractPosition
from ....abs.board import MASTER_BOARD
from ....abs.Rrule import AbstractClueValue
from ....abs.Mrule import AbstractMinesValue

if TYPE_CHECKING:
    ...


def get_value(pos=None, code=None):
    from minesweepervariants.impl.impl_obj import get_value
    return get_value(pos, code)


class Position(AbstractPosition):

    def __repr__(self):
        if self.y > 14:
            return ""
        return f"{'ABCDEFGHIJKLMNO'[self.y]}{self.x + 1}"

    def _up(self, n: int = 1):
        self.x -= 1

    def _down(self, n: int = 1):
        self.x += 1

    def _left(self, n: int = 1):
        self.y -= 1

    def _right(self, n: int = 1):
        self.y += 1

    def _deviation(self, pos: 'AbstractPosition'):
        self.x += pos.x
        self.y += pos.y

    def in_bounds(self, bound_pos: 'AbstractPosition') -> bool:
        return (0 <= self.x <= bound_pos.x and
                0 <= self.y <= bound_pos.y)

    def neighbors(self, *args: int) -> list['Position']:
        """
        按照欧几里得距离从小到大逐层扩散，筛选范围由距离平方控制（不包含当前位置）。

        调用方式（类似 range）：
            neighbors(end_layer)
                返回所有欧几里得距离 ≤ √end_layer 的位置（从第 1 层开始）。
            neighbors(start_layer, end_layer)
                返回所有欧几里得距离 ∈ [√start_layer, √end_layer] 的位置。

        :param args: 一个或两个整数
            - 若提供一个参数 end_layer，视为从 √1 到 √end_layer。
            - 若提供两个参数 start_layer 和 end_layer，视为从 √start_layer 到 √end_layer。
            - 参数非法（数量不为 1 或 2，或值非法）时返回空列表。

        :return: 位置列表，按距离从近到远排序。
        """

        # 解析参数
        if len(args) == 1:
            low, high = 1, args[0]
        elif len(args) == 2:
            low, high = args
        else:
            return []

        # 处理无效参数
        if high < low:
            return []

        x0, y0 = self.x, self.y
        directions = [(dx, dy) for dx in (-1, 0, 1)
                      for dy in (-1, 0, 1) if (dx, dy) != (0, 0)]

        heap = []  # 最小堆存储 (距离平方, x, y)
        visited = {(x0, y0)}
        result = []

        # 处理包含自身的情况 (距离平方=0)
        if low <= 0 <= high:
            result.append(self.clone())

        # 初始化邻居
        for dx, dy in directions:
            nx, ny = x0 + dx, y0 + dy
            d_sq = (nx - x0) ** 2 + (ny - y0) ** 2
            if d_sq <= high:
                heapq.heappush(heap, (d_sq, nx, ny))
                visited.add((nx, ny))

        # 遍历所有可达位置
        while heap:
            d_sq, x, y = heapq.heappop(heap)

            # 检查是否在目标范围内
            if low <= d_sq <= high:
                result.append(Position(x, y, self.board_key))

            # 扩展新位置
            for dx, dy in directions:
                nx, ny = x + dx, y + dy
                if (nx, ny) in visited:
                    continue

                visited.add((nx, ny))
                new_d_sq = (nx - x0) ** 2 + (ny - y0) ** 2

                # 仅考虑距离平方未超过上限的位置
                if new_d_sq <= high:
                    heapq.heappush(heap, (new_d_sq, nx, ny))

        return result


class Board(AbstractBoard):
    name = "Board0"
    version = 0

    __board: list[list[AbstractClueValue | AbstractClueValue | None]]
    __type_board: list[list[str]]
    __dye_board: list[list[bool]]
    variables: list[list[IntVar]] | None = None
    config = {}

    def __init__(self, size: tuple[int, int] = (), code: bytes = None):
        self.model = None
        if code is not None:
            self.size = int((code.split(b"\xff", 1)[0]).decode())
        else:
            self.size = size[0]
        self.__board = [[None for _ in range(self.size)] for _ in range(self.size)]
        self.__type_board = [["N" for _ in range(self.size)] for _ in range(self.size)]
        self.__dye_board = [[False for _ in range(self.size)] for _ in range(self.size)]
        self.config.update({key: False for key in self.CONFIG_FLAGS})
        self.config["row_col"] = True
        self.config["VALUE"] = VALUE_QUESS
        self.config["MINES"] = MINES_TAG
        self.config["interactive"] = True
        if code is None:
            return
        codes = code.split(b"\xff")[1:]
        for pos, _ in self():
            code = codes.pop(0)
            if code[0] == 35:
                self.set_dyed(pos, True)
                code = code[1:]
            if code == b'_':
                continue
            value = get_value(pos, code)
            if value is not None:
                self.set_value(pos, value)
                continue
            raise ValueError(f"unknown type{code}")

    def __call__(
            self, target: Union[str, None] = "always",
            mode: str = "object",
            key: str = MASTER_BOARD
    ) -> Generator[
        Tuple[
            'AbstractPosition',
            Union[
                'AbstractClueValue',
                'AbstractMinesValue',
                str, IntVar, bool, None
            ]],
        Any, None
    ]:
        """
        被调用时循环返回目标值

        @:param
            target (str): 遍历目标类型。可选值：
                - "C": 线索 (Clue)
                - "F": 雷 (Mines)
                - "N": 未定义或未翻开
                - "always": 默认，遍历所有

            mode (str): 返回的目标类型, 可选值:
                - "object":     存储在board内的实例对象
                - "type":       对象的类型('C', 'F', 'N')
                - "variable":   变量对象
                - "dye":        染色bool

        @:return
            当前位置与选择的值。
        """
        if key is None:
            key = MASTER_BOARD
        if key != MASTER_BOARD:
            return
        for posx in range(self.size):
            for posy in range(self.size):
                pos = Position(posx, posy, key)
                pos_type = self.get_type(pos)

                # 检查是否符合目标类型
                if target == "always" or pos_type in target:
                    if mode in "object":
                        yield pos, self.get_value(pos)
                    elif mode == "type":
                        yield pos, pos_type
                    elif mode in "variable":
                        yield pos, self.get_variable(pos)
                    elif mode == "dye":
                        yield pos, self.get_dyed(pos)

    def get_model(self):
        if self.model is None:
            self.model = cp_model.CpModel()
        return self.model

    def has(self, target: str, key=None) -> bool:
        for line in self.__type_board:
            for type_obj in line:
                if type_obj in target:
                    return True
        return False

    def generate_board(self, board_key: str, size: tuple = (), code: bytes = None) -> None:
        get_logger().error("请使用其他版本的题板 该题板不支持副板")
        raise ValueError("ERROR BOARD")

    def boundary(self, key=MASTER_BOARD):
        if key != MASTER_BOARD:
            return None
        return Position(len(self.__board) - 1, len(self.__board[0]) - 1, MASTER_BOARD)

    def encode(self) -> bytes:
        """
        字节头: 尺寸
        无需换行符 初始化自动排序
        '_'表示None
        :return: 字节码
        """
        board_bytes = bytearray()
        board_bytes.extend(f"{self.size}".encode())
        for pos, obj in self():
            board_bytes.extend(b"\xff")
            if self.get_dyed(pos):
                board_bytes.extend(b"#")
            if obj is None:
                board_bytes.extend(b"_")
            else:
                board_bytes.extend(obj.type() + b"|" + obj.code())
        return board_bytes

    def get_type(self, pos: 'AbstractPosition') -> str:
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            return self.__type_board[pos.x][pos.y]
        return ""

    @staticmethod
    def type_value(value) -> str:
        # 查看value的类型
        if isinstance(value, AbstractMinesValue):
            return "F"
        if isinstance(value, AbstractClueValue):
            return "C"
        return "N"

    def get_value(self, pos: 'AbstractPosition') -> Union['AbstractClueValue', 'AbstractMinesValue', None]:
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            return self.__board[pos.x][pos.y]
        return None

    def set_value(self, pos: 'AbstractPosition', value):
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            self.__board[pos.x][pos.y] = value
            self.__type_board[pos.x][pos.y] = self.type_value(value)

    def get_dyed(self, pos: 'AbstractPosition') -> bool | None:
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            return self.__dye_board[pos.x][pos.y]

    def set_dyed(self, pos: 'AbstractPosition', dyed: bool):
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            self.__dye_board[pos.x][pos.y] = dyed

    def get_variable(self, pos: 'AbstractPosition') -> IntVar | None:
        if self.model is None:
            model = self.get_model()
            self.variables = \
                [[model.NewBoolVar(f"{x}.{y}")
                  for y in range(self.size)]
                 for x in range(self.size)]
        if 0 <= pos.x < self.size and 0 <= pos.y < self.size:
            return self.variables[pos.x][pos.y]

    def clear_variable(self):
        if self.variables is None:
            return
        for index_a in range(self.size - 1, -1, -1):
            for index_b in range(self.size - 1, -1, -1):
                del self.variables[index_a][index_b]
            del self.variables[index_a]
        self.variables = None
        self.model = None

    def get_row_pos(self, pos: 'AbstractPosition') -> List["AbstractPosition"]:
        _pos = pos.clone()
        pos_list = [_pos]
        while True:
            _pos = _pos.left()
            if not self.in_bounds(_pos):
                break
            pos_list.append(_pos)
        _pos = pos.clone()
        pos_list = pos_list[::-1]
        while True:
            _pos = _pos.right()
            if not self.in_bounds(_pos):
                break
            pos_list.append(_pos)
        return pos_list

    def get_col_pos(self, pos: 'AbstractPosition') -> List["AbstractPosition"]:
        _pos = pos.clone()
        pos_list = [_pos]
        while True:
            _pos = _pos.up()
            if not self.in_bounds(_pos):
                break
            pos_list.append(_pos)
        _pos = pos.clone()
        pos_list = pos_list[::-1]
        while True:
            _pos = _pos.down()
            if not self.in_bounds(_pos):
                break
            pos_list.append(_pos)
        return pos_list

    def get_pos(self, x: int, y: int, key=MASTER_BOARD) -> Union['AbstractPosition', None]:
        if -self.size < x < self.size and -self.size < y < self.size:
            x = x if x >= 0 else self.size + x
            y = y if y >= 0 else self.size + y
            return Position(x, y, MASTER_BOARD)
        return None

    def get_pos_box(self, pos1: "AbstractPosition", pos2: "AbstractPosition") -> List["AbstractPosition"]:
        if not (self.in_bounds(pos1) and self.in_bounds(pos2)):
            return []
        x_min, x_max = sorted([pos1.x, pos2.x])
        y_min, y_max = sorted([pos1.y, pos2.y])

        result = []
        for y in range(y_min, y_max + 1):
            for x in range(x_min, x_max + 1):
                result.append(self.get_pos(x, y))
        return result

    def batch(self, positions: List['AbstractPosition'],
              mode: str, drop_none: bool = False) -> List[Any]:
        result = []
        for pos in positions:
            if drop_none and not self.in_bounds(pos):
                continue
            if mode == "object":
                result.append(self.get_value(pos))
            elif mode == "variable":
                result.append(self.get_variable(pos))
            elif mode == "type":
                result.append(self.get_type(pos))
            elif mode == "dye":
                result.append(self.get_dyed(pos))
            else:
                raise ValueError(f"Unsupported mode: {mode}")
        return result

    def clear_board(self):
        self.__board = [[None for _ in range(self.size)] for _ in range(self.size)]
        self.__type_board = [["N" for _ in range(self.size)] for _ in range(self.size)]
        self.variables = None

    def show_board(self, show_tag: bool = False):
        r = ""
        for i in range(self.size):
            for j in range(self.size):
                value = self.__board[i][j]
                if value is None:
                    r += "_____" if show_tag else "___"
                else:
                    r += str(value) + ("_" + value.type().decode() if show_tag else "")
                r += "\t"
            r += "\n"
        return r

    def set_config(self, board_key: str, config_name: str, value: bool):
        if board_key != MASTER_BOARD:
            return
        self.config[config_name] = value

    def get_config(self, board_key: str, config_name: str):
        if board_key != MASTER_BOARD:
            return
        return self.config[config_name]

    def get_board_keys(self) -> list[str]:
        return [MASTER_BOARD]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/07 12:50
# @Author  : Wu_RH
# @FileName: config.py.py

DEFAULT_CONFIG = {
    # 图片生成默认参数
    "output_file": "output",       # 默认的img文件名(不含后缀)
    "cell_size": 100,              # 一个单元格的尺寸
    "white_base": False,           # 默认黑底
    "board_class": None,           # 调用默认版本最高的board_class
    "image_debug": False,          # 生成图片时启用debug显示

    # 题板生成默认参数
    "seed": -1,                    # 随机种子
    "logger_lv": "INFO",           # 默认logger为info级别
    "attempts": -1,                # 默认无限次尝试（无-q参数默认20次）
    "query": -1,                   # 如果非-1，则启用-1参数
    "total": -1,                   # 默认自动计算总雷数数量
    "dye": "",                     # 默认无染色
    "used_r": False,               # 默认不启用R推理

    "output_path": ".\\output",   # 保存路径，默认为工作目录下的output
    "timeout": 0,                 # 求解器超时时间（秒），0为无限制
    "workes_number": 5           # 多线程数量
}

IMAGE_CONFIG = config = {
    # 背景色设置
    "background": {
        "white": "#FFFFFF",
        "black": "#000000"
    },

    # 网格线设置
    "grid_line": {
        "white_bg": "#000000",
        "black_bg": "#FFFFFF",
        "width": 0.032
    },

    # 文本设置
    "text": {
        "black": "#FFFFFF",
        "white": "#000000",
        "anchor": "mm"
    },

    # 染色区域
    "dye": {
        "white_bg": "#B3B3B3",
        "black_bg": "#4C4C4C"
    },

    # 描边颜色
    "stroke": {
        "white_bg": "#808080",
        "black_bg": "#D3D3D3"
    },

    # 位置标签（例如 X=E）
    "pos_label": {
        # 标签X=E设置
        "white_bg": "#808080",
        "black_bg": "#808080",
        "size": 0.25
    },

    "assets": "assets",

    # 字体设置
    "font": {
        "name": "CopperplateCC-Heavy.ttf"
    },

    # 边距设置
    "margin": {
        "top_left_right_ratio": 0.7,
        "bottom_ratio": 0.7
    },

    # 坐标轴标签字体比例
    "axis_label": {
        "font_ratio": 0.5
    },

    # 角标尺寸
    "corner": {
        "mini_font_ratio": 0.23
    }
}

PUZZLE_CONFIG = {
    'delimiter': ':'    # 规则键入data的分割符
}



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/08/07 23:45
# @Author  : Wu_RH
# @FileName: -1F.py
"""
[*1F]翻转(Flip): 线索表示其以题板左上角到右下角直线为对称轴的镜像位置周围八格的雷数，不包括自身
"""
from typing import List

from minesweepervariants.abs.Rrule import AbstractClueValue, AbstractClueRule
from minesweepervariants.abs.board import AbstractPosition, AbstractBoard
from minesweepervariants.impl.summon.solver import Switch


class Rulex1F(AbstractClueRule):
    name = ["*1F", "翻转", "Filp"]
    doc = "线索表示其以题板左上角到右下角直线为对称轴的镜像位置周围八格的雷数，不包括自身"

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        for pos, _ in board("N"):
            _pos = pos.clone()
            _pos.x, _pos.y = _pos.y, _pos.x
            value = board.batch(_pos.neighbors(2), mode="type").count("F")
            board.set_value(pos, Valuex1F(pos, code=bytes([value])))
        return board


class Valuex1F(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.pos = pos
        self.value = code[0]

    def __repr__(self):
        return str(self.value)

    def code(self) -> bytes:
        return bytes([self.value])

    def high_light(self, board: 'AbstractBoard') -> List['AbstractPosition'] | None:
        pos = self.pos.clone()
        pos.x, pos.y = pos.y, pos.x
        return pos.neighbors(2)

    @classmethod
    def type(cls) -> bytes:
        return Rulex1F.name[0].encode("ascii")

    def create_constraints(self, board: 'AbstractBoard', switch: 'Switch'):
        model = board.get_model()
        s = switch.get(model, self)
        pos = self.pos.clone()
        pos.x, pos.y = pos.y, pos.x
        var_list = board.batch(pos.neighbors(2), mode="variable", drop_none=True)
        model.Add(sum(var_list) == self.value).OnlyEnforceIf(s)




#!/usr/bin/env python3
# -*- coding:utf-8 -*-
#
# @Time    : 2025/07/09 07:14
# @Author  : Wu_RH
# @FileName: 2L.py
"""
[2L] 误差 (Liar)：每行每列恰有一个误差线索。误差线索的值比真实值大 1 或小 1 [副版规则]
"""
from ....abs.Rrule import AbstractClueRule, AbstractClueValue
from ....abs.board import AbstractBoard, AbstractPosition
from ....utils.impl_obj import VALUE_CIRCLE, VALUE_CROSS
from ....utils.tool import get_random, get_logger

NAME_2L = "2L"


def select(matrix: list[list[bool]]) -> list[tuple[int, int]]:
    rnd = get_random()
    n = len(matrix)

    graph = [[] for _ in range(n)]
    for i in range(n):
        for j in range(n):
            if matrix[i][j]:  # False 是可选项
                graph[i].append(j)

    for g in graph:
        rnd.shuffle(g)

    match_to = [-1] * n

    def dfs(_u: int, _visited: list[bool]) -> bool:
        for v in graph[_u]:
            if not _visited[v]:
                _visited[v] = True
                if match_to[v] == -1 or dfs(match_to[v], _visited):
                    match_to[v] = _u
                    return True
        return False

    for u in range(n):
        visited = [False] * n
        if not dfs(u, visited):
            return []

    return [(match_to[j], j) for j in range(n)]


class Rule2L(AbstractClueRule):
    name = ["2L", "误差", "Liar"]
    doc = "每行每列恰有一个误差线索。误差线索的值比真实值大 1 或小 1"

    def __init__(self, board: "AbstractBoard" = None, data=None) -> None:
        super().__init__(board, data)
        bound = board.boundary()
        if bound.x != bound.y:
            raise ValueError("请输入一个正方形题板")
        for key in board.get_interactive_keys():
            _bound = board.boundary(key)
            if (_bound.x != bound.x and
                    _bound.y != bound.y):
                raise ValueError("请保证其他题板尺寸均一致")
        board.generate_board(NAME_2L, (bound.x + 1, bound.y + 1))

    def fill(self, board: 'AbstractBoard') -> 'AbstractBoard':
        random = get_random()
        logger = get_logger()
        boundary = board.boundary()
        keys = board.get_interactive_keys()

        matrix = []
        for x in range(boundary.x + 1):
            row = []
            for y in range(boundary.y + 1):
                row.append(all(
                    board.get_type(board.get_pos(x, y, key)) != "F"
                    for key in keys
                ))
            matrix.append(row)

        pos_map = select(matrix)

        if not pos_map:
            raise ValueError("[2L]: has a line is all mines")

        for pos, _ in board("N"):
            value = board.batch(pos.neighbors(2), mode="type").count("F")
            if (pos.x, pos.y) in pos_map:
                value = random.choice([value + 1, value - 1])
                if value == -1:
                    value = 1
                if value == 9:
                    value = 7
                obj = Value2L(pos, bytes([random.choice([value])]))
            else:
                obj = Value2L(pos, bytes([value]))
            board.set_value(pos, obj)
            logger.debug(f"[2L]put {obj} to {pos}")

        for pos, _ in board(key=NAME_2L):
            if (pos.x, pos.y) in pos_map:
                board.set_value(pos, VALUE_CIRCLE)
                logger.debug(f"[2L]put O to {pos}")
            else:
                board.set_value(pos, VALUE_CROSS)
                logger.debug(f"[2L]put X to {pos}")

        return board

    def init_clear(self, board: 'AbstractBoard'):
        for pos, _ in board(key=NAME_2L):
            board.set_value(pos, None)

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        bound = board.boundary(key=NAME_2L)

        row = board.get_row_pos(bound)
        for pos in row:
            line = board.get_col_pos(pos)
            line_var = board.batch(line, mode="variable", drop_none=True)
            model.Add(sum(line_var) == 1).OnlyEnforceIf(s)

        col = board.get_col_pos(bound)
        for pos in col:
            line = board.get_row_pos(pos)
            line_var = board.batch(line, mode="variable", drop_none=True)
            model.Add(sum(line_var) == 1).OnlyEnforceIf(s)

        for pos, _ in board(key=NAME_2L):
            _pos = pos.clone()
            var = board.get_variable(pos)
            for key in board.get_interactive_keys():
                _pos.board_key = key
                key_var = board.get_variable(_pos)
                model.Add(key_var == 0).OnlyEnforceIf([var, s])


class Value2L(AbstractClueValue):
    def __init__(self, pos: 'AbstractPosition', code: bytes = b''):
        self.value = code[0]
        self.nei = pos.neighbors(2)
        self.pos = pos.clone()
        self.pos.board_key = NAME_2L

    def __repr__(self) -> str:
        return str(self.value)

    @classmethod
    def method_choose(cls) -> int:
        return 1

    @classmethod
    def type(cls) -> bytes:
        return Rule2L.name[0].encode("ascii")

    def code(self) -> bytes:
        return bytes([self.value])

    def create_constraints(self, board: 'AbstractBoard', switch):
        model = board.get_model()
        s = switch.get(model, self)
        nei_vars = board.batch(self.nei, mode="variable", drop_none=True)
        var = board.get_variable(self.pos)

        model.Add(sum(nei_vars) == self.value).OnlyEnforceIf(var.Not())

        tmp_a = model.NewBoolVar("tmp_a")
        tmp_b = model.NewBoolVar("tmp_b")

        model.Add(sum(nei_vars) == self.value + 1).OnlyEnforceIf([tmp_a, var, s])
        model.Add(sum(nei_vars) == self.value - 1).OnlyEnforceIf([tmp_b, var, s])

        model.AddBoolOr([tmp_a, tmp_b]).OnlyEnforceIf(s)
